<?
/* --------------------------------------------------------------------
   Copyright (c) 2008-2012 Daniel Pedemont (dan@pedemont.net)
   -------------------------------------------------------------------- */
/**
 * HTTP style listener as IPC plugin module
 *
 * web interface plugin for the daemon IPC eg. reset services, debug, shutdown
 *
 * @link http://dev.pedemont.com/sonic/
 * @link http://dev.pedemont.com/sonic/downloads/src/lib/IPC/sonic.IPC.http.php
 * @package sonic_IPC
 * @version 0.1.5
 * @abstract an IPC plugin module
 */

define('SONIC_IPC_HTTP_VERSION', '0.1.6');

/**
 * IPC main
 *
 * @param array &$opts option flags, (array) argv, (boolean) quiet, (integer) verbose, (array) config, (array) services, (string) buffer, (boolean) finished, (string) output, (boolean) services_reset
 * @return mixed false on failure or NULL on success
*/
function sonic_daemon_IPC_http_main(&$opts)
{

//TODO: none of this shite is checking to see what mode we are in
//	ie. might not need to exec threaded schtuff

  $_sd_pid = getmypid();

  $argv = $opts['argv'];
  $quiet = $opts['quiet'];
  $verbose = $opts['verbose'];
  $services = $opts['services'];
  $config = &$opts['config'];

  $buf = $opts['buffer'];
  $finished = &$opts['finished'];
  $out = &$opts['output'];
  $services_reset = &$opts['services_reset'];

  $services_reset = false;
  $finished = false;

  $_h_buf_a = array();
  $_h_buf_a = sonic_header2array($buf);

  if ($method);
  else
  {
    if (preg_match('/^\s*GET/im', $buf)) { $method = 'get'; }
    else if (preg_match('/^\s*POST/im', $buf)) { $method = 'post'; }
    else
    {
      $out = '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
            .'<HTML><HEAD>'."\n"
            .'<TITLE>501 Not Implemented</TITLE>'."\n"
            .'</HEAD><BODY>'."\n"
            .'<H1>Bad Request</H1>'."\n"
            .'Your browser sent a request that this server could not understand.<P>'."\n"
            .'</BODY></HTML>'."\n";

      if ($quiet);
      else
        echo '! ['.$_sd_pid.'] '.sonic_datef()
            .' [SONICD_IPC] [HRC:501] Error: Not Implemented '
            .base64_encode($buf)."\n";

      return false;
    }
  }

  if ($method == 'get')
  {
//    if (preg_match('/\r{0,1}\n\r{0,1}\n$/', $buf));
  }
  else if ($method == 'post')
  {
    if (preg_match('/\r{0,1}\n\r{0,1}\n/', $buf))//, $matches))
    {
      list($_h_buf_h, $_h_buf_b) = preg_split('/\r{0,1}\n\r{0,1}\n/', $buf, 2);
      if ($_h_buf_a['Content-Length'])
      {
        if (($_h_buf_a['Content-Length'] + 0) > strlen($_h_buf_b));
        else
        {
          $out = 'post probably still in progress';
          return false;
        }
      }
      else
      {
        $out = 'this is probably an error';
        return false;
      }
    }
  }
  unset($_h_buf_a);

  $finished = true;

  $header_raw = $buf;
  $header = sonic_header2array($buf);
  unset($buf);

  if (($_she = sonic_http_env_get($header_raw, $_she_globs)) > 0);
  else if ($_she != -3 && $_she != -5) //404 "Not Found" handled below
  {
//sonic_http_env()
// 1:   OK
// 0:   not get|post aka unknown method //405 "Method Not Allowed"
//-1:   bad get|post aka invalid syntax //400 "Bad Request"
//-2:   bad post			//417 "Expectation Failed"
//-3:   script filename does not exist  //404 !handled out of this scope
//-4:   bad script aka invalid syntax   //400 "Invalid Syntax"
//-5:   no default index found		//404 !handled out of this scope

    switch ($_she)
    {
      case  0:
        $hrc = 405;
        $_she_msg = 'Method Not Allowed';
      break;
      case -2:
        $hrc = 417;
        $_she_msg = 'Expectation Failed';
      break;
      case -4:
        $hrc = 400;
        $_she_msg = 'Invalid Syntax';
      break;

      case -1:
      default:
        $hrc = 400;
        $_she_msg = 'Bad Request';
    }

    $out = ''
	.'<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">'."\n"
	.'<HTML><HEAD>'."\n"
	.'<TITLE>400 '.$_she_msg.'</TITLE>'."\n"
	.'</HEAD><BODY>'."\n"
	.'<H1>'.$_she_msg.'</H1>'."\n"
	.'Your browser sent a request that this server could not understand.<P>'."\n"
	.'</BODY></HTML>'."\n"
	.'';

    if ($quiet);
    else
      echo '! ['.$_sd_pid.'] '.sonic_datef()
          .' [SONICD_IPC] [HRC:'.$hrc.'] Error: '.$_she_msg."\n";

    return false;
  }
  unset($_she);

  $hrc = 200;
  $out_headers = array('Content-Type' => 'text/html');

//MAIN

  $op = $_she_globs['_REQUEST']['op'];
  $reset = $_she_globs['_REQUEST']['reset'];
  $name = $_she_globs['_REQUEST']['name'];

  switch ($op)
  {
    case 'IPC':
    case 'IPC-session':
    case 'IPC-debug':
    break;

    default:
      $finished = false;
      return false;
  }

  if ($op == 'IPC-session')
  {
    global	$_sd_ipc_sessions;

    $name = strtolower($name);
    switch ($name)
    {
      case 'set':
      case 'get':
        $name = strtolower($name);
      break;

      default:
//        if ($_she_globs['_SERVER']['HTTP_ACCEPT'])
//          $name = 'help';
//        else
          $name = 'ack';
    }

    $key = $_she_globs['_REQUEST']['key'];

    $var_type = $_she_globs['_REQUEST']['type'];

    switch ($var_type)
    {
      case 'array':
      case 'integer':
      case 'string':
      case 'boolean':
      break;

      default:
        $var_type = 'array';
    }

    if ($key == '')
    {
      $_out = 'Error: key required';
    }
    else if ($name == 'ack')
    {
      $_out = 'OK';
    }
    else if ($name == 'get')
    {
      if ($slp = $_she_globs['_REQUEST']['lock_pass'])
      {
        $lock_check = true;
      }
      else
      {
        $slp = false;
        $lock_check = false;
      }

      //$slp set to false on lock failure during get request
      $var = sonic_daemon_IPC_session($slp, $key, null, false);

      $var_type = gettype($var);

      switch ($var_type)
      {
        case 'NULL':
          $_out = NULL;
        break;

        case 'integer':
        case 'boolean':
        case 'string':
          $_out = $var;
        break;

        case 'array':
          $_out = sonic_array2str($var);
        break;

        default:
          $var_type = 'string';
          $_out = var_export($var, true);
      }

      $out_headers['Sonic-IPC-Variable-Type'] = $var_type;

      if ($slp)
        $out_headers['Sonic-IPC-Lock-Pass'] = $slp;

      if ($lock_check && $slp) {}
      else if ($lock_check)
      {
        $out_headers['Sonic-IPC-Lock-Failure'] = 'True';
      }
    }
    else if ($name == 'set')
    {
      if ($var = $_she_globs['_REQUEST']['var'])
      {
        if ($slp = $_she_globs['_REQUEST']['lock_pass'])
        { 
          $lock_check = true;
        }
        else
        { 
          $slp = false;
          $lock_check = false;
        }

        $var = base64_decode($var);

        switch ($var_type)
        {
          case 'array':
            $var = sonic_str2array($var);
          break;

          case 'boolean':
          case 'integer':
          case 'NULL':
            $var = settype($var, $var_type);
          break;
        }

        $slp_snapshot = $slp;

        //$slp set to current lock pass on locked failure
        if (sonic_daemon_IPC_session($slp, $key, $var, false))
        {
          $_out = 'OK';
        }
        else
        {
          if ($lock_check && $slp_snapshot != $slp)
          {
            $out_headers['Sonic-IPC-Lock-Pass'] = $slp;
            $out_headers['Sonic-IPC-Lock-Failure'] = 'True';

            $_out = 'Error: set failed because its currently locked';
          }
          else
            //safety *shrug* will this be reached ever?
            $_out = 'Error: set failed';
        }
      }
      else
      {
        $_out = 'Error: no var supplied';
      }
    }
  }
  else if ($op == 'IPC')
  {
//    if ($reset == '');
//    else
//      $services_reset = true;


    $name = strtolower($name);
    switch ($name)
    {
      case 'kill':
      case 'commit':
      case 'shutdown':
      case 'convert':
      case 'help':
      case 'ack':
      break;

      default:
        if ($_she_globs['_SERVER']['HTTP_ACCEPT'])
          $name = 'help';
        else
          $name = 'ack';
    }

    if ($name == 'help')
    {
//      $ps_argv = $argv;
      $ps_argv = array();

      $ps_argv['services'] = $config['services'];
      foreach ($services as $service => $service_conf)
      {
        foreach ($config as $k => $v)
        {
          if (preg_match('/^service-'.$service.'-/', $k))
            $ps_argv[$k] = $v;
        }
      }

//TODO: make encode optional? gui looks ugly
      $rt_part_arg = '';
      foreach ($ps_argv as $k => $v)
      {
        if (preg_match('/\s/', $v))
        {
          $v = '\"'.$v.'\"';
          $rt_part_arg .= "\"--$k=".$v."\" \\\n";
        }
        else
          $rt_part_arg .= "--$k=".$v." \\\n";

        
//        $rt_part_arg .= "--$k=".urlencode($v)." \\\n";
//        $rt_part_arg .= "--$k=".$v." \\\n";
      }
      if ($rt_part_arg)
        $rt_part_arg = substr($rt_part_arg, 0, -2);


      global	$_sd_threads, $_sd_threads_service, $_sd_threads_service_active, $_sd_threads_shutdown;

//      $_out_sd_threads = '';
//      foreach ($_sd_threads as $_tpid => $_tinfo)
//      {
//        $_out_sd_threads .= ', '.$_tpid.':'.$_tinfo['name'];
//      }
//      $_out_sd_threads = substr($_out_sd_threads, 2);

      $_out_sd_threads_service = '';
      foreach ($_sd_threads_service as $_tname => $_tpids)
      {
        $_out_sd_threads_service .= ', <b>'.$_tname.'</b>:';
        $_out_sd_threads_service_pids = '';
        foreach ($_tpids as $_tpid => $_tinfo)
        {
          $_out_sd_threads_service_pids .= ', '
					  .($_sd_threads_service_active[$_tname][$_tpid]
					    ?
					      $_tpid 
					    :
					      ' <a href="'.$_she_globs['_SERVER']['SCRIPT_URI']
			                     .'?op=IPC&name=kill&pid='.$_tpid.'&sig=SIGKILL" />'.$_tpid.'</a>'
					   );
        }

        $_out_sd_threads_service_pids = substr($_out_sd_threads_service_pids, 2);
        $_out_sd_threads_service .= $_out_sd_threads_service_pids;
      }
      $_out_sd_threads_service = substr($_out_sd_threads_service, 2);
      unset($_out_sd_threads_service_pids);

      $_out_sd_threads_service_active = '';
      foreach ($_sd_threads_service_active as $_tname => $_tpids)
      {
        $_out_sd_threads_service_active .= ', <b>'.$_tname.'</b>:';
        $_out_sd_threads_service_active_pids = '';
        foreach ($_tpids as $_tpid => $_tinfo)
          $_out_sd_threads_service_active_pids .= ', '.$_tpid;
        $_out_sd_threads_service_active_pids = substr($_out_sd_threads_service_active_pids, 2);
        $_out_sd_threads_service_active .= $_out_sd_threads_service_active_pids;
      }
      $_out_sd_threads_service_active = substr($_out_sd_threads_service_active, 2);
      unset($_out_sd_threads_service_active_pids);

      $_out_thread = '<table cellspacing="0" cellpadding="1" border="0">'."\n"
                    .'<tr><td><font size="-1">Active Services</font></td>'
                    .'  <td><font size="-2">'.$_out_sd_threads_service_active.'</font></td></tr>'."\n"
                    .'<tr><td><font size="-1">All Services</font></td>'
                    .'  <td><font size="-2">'.$_out_sd_threads_service.'</font></td></tr>'."\n"
//                    .'<tr><td><font size="-1">All</font></td>'
//                    .'	<td><font size="-2">'.$_out_sd_threads.'</font></td></tr>'."\n"
                    .'</table>';

      unset($_out_sd_threads_service_active, $_out_sd_threads_service, $_out_sd_threads);

//uptime
      $_ut = time() - SONICD_INIT_TIME;

      $periods = array('second', 'minute', 'hour', 'day', 'week', 'month', 'year', 'decade');
      $lengths = array('60','60','24','7','4.35','12','10');

      if ($_ut)
      {
        $tense = 'ago';
      }
      else if ($_ut == 0)
      {
        $_out_uptime = 'Now';
      }
      else
      {
        $tense = 'from now';
      }

      if ($_ut == 0) { }
      else
      { 
        $j = 0;

        for (;$_ut >= $lengths[$j] && $j < count($lengths) - 1; $j++)
          $_ut /= $lengths[$j];

        $_ut = round($_ut);

        if ($_ut != 1)
          $periods[$j].= 's';

        $_out_uptime = $_ut.' '.$periods[$j].' '.$tense;

        unset($j);
      }
      unset($_ut, $tense, $periods, $lengths);
//uptime

      $_out = ''
	.'<html>'."\n"
	.'<head>'."\n"
	.'<title>Sonic Daemon - IPC - Help</title>'."\n"
	.'</head>'."\n"
	.'<body>'."\n"
	.'<form>'."\n"
	.'<input type="hidden" name="op" value="IPC" />'
	.'<input type="submit" name="name" value="convert" /> '
	.'RUN name '
	.'<input type="radio" name="pass_name" value="commit" checked="checked" /> commit'."\n"
	.'<div style="padding-right: 1em; text-align: center; height: 69%; border: thin solid black;">'."\n"
	.'<p align="right">'
	.'<a href="?op=IPC&name=commit" title="service(s) reset/re-commit will SIGINT active children">@</a> : '
	.'<a href="?op=IPC&name=commit&services=" title="commit service(s) shutdown with an empty service arg \'--services=\'">!</a> : '
	.'<a href="?op=IPC&name=shutdown" title="Ctrl-C the daemon, may require multiple attempts to shutdown the daemon">X</a><br />'
	.'<i>insert <a href="http://dev.pedemont.com/sonic/">server daemon</a> <b>runtime arguments</b> below, click once, wait many.</i><br />'
	.'deamon loaded '.$_out_uptime.'<br />'."\n"
	.'<textarea style="width: 88%; height: 78%;" name="pass_args">'."\n"
	.$rt_part_arg
	.'</textarea>'."\n"
	.'<p style="background: silver;" align="left"><a href="?op=IPC-debug">debug</a>: <font color="gray">environment eg. zombie information</font>'
        .$_out_thread
        .'</p>'
	.'</div>'."\n"
	.'</form>'."\n"
	.'</body>'."\n"
	.'</html>'."\n"
	.'';
    }
    else if ($name == 'convert')
    {
      $pass_name = $_she_globs['_REQUEST']['pass_name'];

      $pass_args_wrk = stripslashes($_she_globs['_REQUEST']['pass_args']);
      $pass_args_wrk = preg_replace('/\s*\\\[\r\n]+/', ' ', $pass_args_wrk);
      $pass_args_wrk = preg_replace('/\s*"(.+[^\\\])"\s*/', ' \1 ', $pass_args_wrk);
      $pass_args_wrk = preg_replace('/\\\"/', '"', $pass_args_wrk);
      $pass_args_wrk = trim($pass_args_wrk);

      $pass_args = preg_split('/\s+/', $pass_args_wrk);
      array_unshift($pass_args, __FILE__);
      $pass_args = sonic_args($pass_args);

      $pass_URL_pa = '';
      foreach ($pass_args as $_kav => $_vav)
        $pass_URL_pa .= '&'.$_kav.'='.urlencode($_vav);

      $pass_URL = $_she_globs['_SERVER']['SCRIPT_URI'].'?op=IPC&name='.$pass_name.$pass_URL_pa;

//var_dump($pass_URL_pa, $pass_name, $pass_URL);

//old
//      //--key=value -k=v
//
//      $pass_args = stripslashes($_she_globs['_REQUEST']['pass_args']);
//
//      $pass_args = preg_replace('/\s*--([^=]+)\s*=\s*(.+)/U', '&\\1=\\2', $pass_args);
//      $pass_args = preg_replace('/\s*-([^-]{1,1})\s+((?!-).+)[\s\\\]*/U', '&\\1=\\2', $pass_args);
//
//      //--flag -f
//      $pass_args = preg_replace('/^--([^=]+)\s*/', '&\\1=true', $pass_args);
//      $pass_args = preg_replace('/\s+--([^=]+)$/', '&\\1=true', $pass_args);
//      $pass_args = preg_replace('/\s+--([^=]+)\s+/', '&\\1=true', $pass_args);
//      //
//      $pass_args = preg_replace('/^-([^-]{1,1})\s*/', '&\\1=true', $pass_args);
//      $pass_args = preg_replace('/\s+-([^-]{1,1})$/', '&\\1=true', $pass_args);
//      $pass_args = preg_replace('/\s+-([^-]{1,1})\s+/', '&\\1=true', $pass_args);
//
//      //new line fix
//      $pass_args = preg_replace('/[\s*\\\]*&/', '&', $pass_args);
//      $pass_args = preg_replace('/[\s*\\\]*$/', '', $pass_args);
//
////      if ($pass_args[0] == '&')
////        $pass_args[0] = '?';
//
//      $pass_URL = $_she_globs['_SERVER']['SCRIPT_URI']
//		.'?op=IPC&name='.$pass_name.$pass_args;


      if ($_she_globs['_SERVER']['HTTP_ACCEPT'])
      {
        $_out = ''
	.'<html>'."\n"
	.'<head>'."\n"
	.'<title>Sonic Daemon - IPC - pre-COMMIT</title>'."\n"
	.'</head>'."\n"
	.'<body>'."\n"
	.'<a href=\''.$pass_URL.'\'>CONFIRM RUN</a>'
	.'<div style="padding: 1em; text-align: center; height: 69%; border: thin solid black;">'."\n"
//	.'<p align="right"><font size="-1"><a href=\''.$pass_URL.'\'>'.substr($pass_args, 1).'</a></font></p>'
	.'<p align="right"><font size="-1"><a href=\''.$pass_URL.'\'>'.$pass_args_wrk.'</a></font></p>'
	.'<br />'."\n"
	.'<textarea style="width: 88%; height: 78%;" name="pass_args">'."\n"
	.stripslashes($_she_globs["_REQUEST"]['pass_args'])
	.'</textarea>'."\n"
	.'<p style="background: silver;" align="left"><a href="?op=IPC&name=help">help</a> : <a href="?op=IPC-debug">debug</a></p>'
	.'<p style="border: thin solid silver;" align="right"><a href="?op=IPC&name=shutdown">shutdown</p>'
	.'</div>'."\n"
	.'</body>'."\n"
	.'</html>'."\n"
	.'';

//$_out = $pass_args;
      }
      else
      {
        $out_headers['Content-Type'] = 'text/plain';
        $out_headers['Location'] = $pass_URL;

        $hrc = 302;

        $_out = 'OK';
      }
    }
    else if ($name == 'shutdown')
    {
      sonic_daemon_thread_shutdown();

      $out_headers['Content-Type'] = 'text/plain';
      $out_headers['Location'] = $_she_globs['_SERVER']['SCRIPT_URI'].'?op=IPC';

      $hrc = 302;

      $_out = 'OK';
    }
    else if ($name == 'commit')
    {
      $_out =''
        .'<html>'."\n"
        .'<head>'."\n"
        .'<title>Sonic Daemon - IPC - COMMIT</title>'."\n"
        .'</head>'."\n"
        .'<body>'."\n"
	.'<center><a href="?op=IPC&name=help">help</a>&nbsp;:&nbsp;<a href="?op=IPC-debug">debug</a></center>';

      $_out .= '<b>before</b>:<br />'.str_replace("\n", '<br />', var_export($config, true)).'<br />';

      $_sgr_gt = gettype($_she_globs['_REQUEST']);
      $_sgc_gt = gettype($_she_globs['_COOKIE']);

      if ($_sgr_gt == 'array' && $_sgc_gt == 'array')
      {
        $argv = array_diff($_she_globs['_REQUEST'], $_she_globs['_COOKIE']);
      }
      else if ($_sgr_gt == 'array')
      {
        $argv = &$_she_globs['_REQUEST'];
      }
      else
        $argv = array();

/*
      unset($argv['op'], $argv['name']);

      foreach ($argv as 

//var_dump($argv);

//      $argv = sonic_args($argv);

//var_dump($argv);

$argv = array();
*/

      foreach ($argv as $k => $v)
      {
        if (preg_match('/^[0-9]+$/', $v))
          $v += 0;

        if (preg_match('/^service-([^-\s]+)-[^\s]+$/i', $k, $matches))
          $config[$k] = $v;

        if (preg_match('/^s-([^-\s]+)-([^\s]+)$/i', $k, $matches))
          $config['service-'.$matches[1].'-'.$matches[2]] = $v;

        if (preg_match('/^services$/i', $k, $matches))
          $config[$k] = $v;

        if (preg_match('/^S$/i', $k, $matches))
          $config['services'] = $v;
      }

      $_out .= '<b>after</b>:<br />'.str_replace("\n", '<br />', var_export($config, true)).'<br />'
	.'</body>'."\n"
	.'</html>'."\n"
	.'';

      $services_reset = true;
    }
    else if ($name == 'kill')
    {
      $pid = $_she_globs['_REQUEST']['pid'] + 0;

      if ($sig = $_she_globs['_REQUEST']['sig'])
      {
        if ($_sig = @constant($sig))
          $sig = $_sig;
        unset($_sig);

        $sig += 0;

        //lazy/expensive thread/rotate mode check
        //they shouldnt be here otherwise anyway *shrug*
        if (function_exists(sonic_daemon_thread_info))
          $zombie_pids = sonic_daemon_thread_info(array('quiet' => $quiet, 'verbose' => $verbose));
        else
          $zombie_pids = array();

        if (in_array($pid, array_keys($zombie_pids)))
        {
          if (posix_kill($pid, $sig))
            $_out = 'OK';
          else
            $_out = 'FAIL';
        }
        else
          $_out = 'FORBIDDEN OR NON EXISTANT';

        unset($zombie_pids);
      }
      else
        $_out = 'ERROR';

      $out_headers['Content-Type'] = 'text/plain';

      unset($pid, $sig);
    }
    else if ($name == 'ack')
    {
      $out_headers['Content-Type'] = 'text/plain';
      $_out = 'OK';
    }
    else
    {
      $out_headers['Content-Type'] = 'text/plain';
      $_out = 'NYA';
    }
  }
  else if ($op == 'IPC-debug')
  {
    $_sd_ti = var_export(sonic_daemon_thread_info(array('quiet' => $quiet, 'verbose' => $verbose)), true);
    $_sd_ti_rep = ' <a href="'
		 .$_she_globs['_SERVER']['SCRIPT_URI']
		 .'?op=IPC&name=kill&pid=\\1&sig=SIGKILL" />\\1</a> =&gt; \\2';
    $_sd_ti = preg_replace('/\s+([0-9]+)\s+=>\s+([^\n\r]+)[\n\r]+/', $_sd_ti_rep, $_sd_ti);
    $_sd_ti = preg_replace('/[\n\r]+/', '<br />', $_sd_ti);

//NB, var_export() may cause; PHP Fatal error:  Nesting level too deep - recursive dependency?
//eg. [GLOBALS] => Array *RECURSION*

    $_out = ''
	.'<html>'."\n"
	.'<head>'."\n"
	.'<title>Sonic Daemon - IPC - DEBUG</title>'."\n"
	.'</head>'."\n"
	.'<body>'."\n"
	.'<b>globals FETCH:</b>'."\n"
	.'<xmp>'."\n"
	.print_r($_she_globs, true)."\n"
	.'</xmp>'."\n"
	.'<b>globals ACTUAL:</b>'."\n"
	.'<xmp>'."\n"
	.print_r($GLOBALS, true)."\n"
	.'</xmp>'."\n"
	.'<b>Sonic ZOMBIE services:</b>'."<br />\n"
	.'<br />'."\n"
	.'<div>'."\n"
	.$_sd_ti
	.'</div>'."\n"
	.'<br />'."\n"
	.'<b>Sonic DAEMON ARGS:</b>'."\n"
	.'<xmp>'."\n"
	.print_r($argv, true)."\n"
	.'</xmp>'."\n"
	.'<b>Sonic DAEMON CONFIG:</b>'."\n"
	.'<xmp>'."\n"
	.print_r($config, true)."\n"
	.'</xmp>'."\n"
	.'<b>Sonic DAEMON SERVICES:</b>'."\n"
	.'<xmp>'."\n"
	.print_r($services, true)."\n"
	.'</xmp>'."\n"
	.'</body>'."\n"
	.'</html>'."\n"
	.'';
        unset($_sd_ti);
  }
  else
    $_out = "IPC http 'UMmm....'";

//MAIN

//NB, !$+ log macro
  if ($quiet);
  else
    echo '! ['.$_sd_pid.'] '.sonic_datef()
        .' [SONICD_IPC] [HRC:'.$hrc.'] '.$_she_globs['_SERVER']['REQUEST_URI']."\n";

  $out = sonic_http_response($hrc, $_out, $out_headers);
}

?>
