<?
/* --------------------------------------------------------------------
   Copyright (c) 2004-2012 Daniel Pedemont (dan@pedemont.net)
   -------------------------------------------------------------------- */

function sonic_daemon_main()
{
  error_reporting(E_ALL ^ E_NOTICE);

//SCOPE PROTECTION (variables only, php doesnt protect functions)
  static $_sd_exec_main_lock = false;
  if ($_sd_exec_main_lock)
  {
    echo '! [SONICD] '.sonic_datef().' Notice: main(): LOCKED and currently already running'."\n";
    return;
  }
  else
    $_sd_exec_main_lock = true;

//MAIN

define('SONICD_INIT_TIME', time());
define('SONICD_VERSION', '0.2.39');
define('SONIC_IPC_BYTE_SIZE', 2048);

//PCNTL check
if (extension_loaded('pcntl') && extension_loaded('posix'))
  define('SONICD_PCNTL', true);
else
  define('SONICD_PCNTL', false);

if (extension_loaded('sysvsem') && extension_loaded('sysvshm'))
  define('SONICD_SYSV', true);
else
  define('SONICD_SYSV', false);

function sonic_daemon_load_config_IPC($config = array(), $opts = array('fatal' => false, 'quiet' => true))
{
//  $pcntl_loaded = (extension_loaded('pcntl') && extension_loaded('posix'));
  $pcntl_loaded = SONICD_PCNTL;
  $_sd_pid = $pcntl_loaded ? posix_getpid() : getmypid();

  $fatal = &$opts['fatal'];
  $quiet = &$opts['quiet'];

  if ($sd_ipc = &$config['ipc_enable']);
  else
  {
    if (!$quiet) echo '* IPC disabled'."\n";
    return false;
  }

  $sd_ipc = array();

  $sd_ipc['wait_time'] = &$config['ipc_wait_time'];
  $sd_ipc['wait_time'] += 0;
  $sd_ipc['default_msg'] = &$config['ipc_default_msg'];

  if ($sd_ipc['handlers'] = &$config['ipc_handlers'])
  {
    $config_handlers = preg_split('/\s*,\s*/', $sd_ipc['handlers']);
    $sd_ipc['handlers_array'] = array();
  }
  else if ($fatal)
  {
    die('! ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] Error: no handlers are set to load'."\n");
  }
  else
    return false;

  //load stream parent config
  while (list($i, $_sd_IPC_name) = each($config_handlers))
  {
    if ($quiet);
    else
      echo '+ reading IPC handler information for ['.$_sd_IPC_name.']'."\n";

    $_sd_IPC = $_SERVER[SONIC_LIB].DIRECTORY_SEPARATOR.'IPC'.DIRECTORY_SEPARATOR.'sonic.IPC.'.$_sd_IPC_name.'.php';

    if (file_exists($_sd_IPC))
    { 
      require_once($_sd_IPC);
    }
    else if ($fatal)
    { 
      die('! ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] Error: can not load '.($_sd_IPC_name ? $_sd_IPC_name.' ' : '').'handler'."\n");
    }
    else
      return false;

    array_push($sd_ipc['handlers_array'], $_sd_IPC_name);
  }

  //helper suff. (setup for inbound socket type stream)
  if (isset($config['ipc_socket_domain']))
  {
    $sd_ipc['socket_domain'] = &$config['ipc_socket_domain'];

    if ($_def = @constant($sd_ipc['socket_domain']))
      $sd_ipc['socket_domain'] = $_def;

    $sd_ipc['socket_domain'] += 0;
  }
  else
    $sd_ipc['socket_domain'] = SONIC_SOCK_DOMAIN;

  if (isset($config['ipc_socket_type']))
  {
    $sd_ipc['socket_type'] = &$config['ipc_socket_type'];

    if ($_def = @constant($sd_ipc['socket_type']))
      $sd_ipc['socket_type'] = $_def;

    $sd_ipc['socket_type'] += 0;
  }
  else
    $sd_ipc['socket_type'] = SONIC_SOCK_TYPE;

  if (isset($config['ipc_socket_protocol']))
  {
    $sd_ipc['socket_protocol'] = &$config['ipc_socket_protocol'];

    if ($_def = @constant($sd_ipc['socket_protocol']))
      $sd_ipc['socket_protocol'] = $_def;

    $sd_ipc['socket_protocol'] += 0;
  }
  else
    $sd_ipc['socket_protocol'] = SONIC_SOCK_PROTOCOL;

  if (($sd_ipc['socket_options'] = &$config['ipc_socket_options']) != NULL);
  else
    $sd_ipc['socket_options'] = SONIC_SOCK_OPTIONS;

  //help ness.
  $sd_ipc['port'] = &$config['ipc_port'];
  $sd_ipc['port'] += 0;

  if ($sd_ipc['ip'] = &$config['ipc_ip']);
  else
  {
    switch ($sd_ipc['socket_domain'])
    {
      case AF_INET:
        $sd_ipc['ip'] = '0.0.0.0';
      break;
      case AF_INET6:
        $sd_ipc['ip'] = 'ff:ff:ff:ff:ff:ff';
      break;
      case AF_UNIX: //port = 0
        $sd_ipc['ip'] = SONIC_TMP_PATH.DIRECTORY_SEPARATOR.'sonic_IPC_'.md5(time()).'.sock';
        $sd_ipc['port'] = 0;
      break;
    }
  }

  return $sd_ipc;
}

function sonic_daemon_load_IPC(&$sd_ipc, $opts = array('fatal' => false, 'quiet' => true))
{
//  $pcntl_loaded = (extension_loaded('pcntl') && extension_loaded('posix'));
  $pcntl_loaded = SONICD_PCNTL;
  $_sd_pid = $pcntl_loaded ? posix_getpid() : getmypid();

  $fatal = $opts['fatal'];
  $quiet = $opts['quiet'];

  if (isset($sd_ipc['port'])) //socket setup trigger, check above
  {
//word map tcp -> SOL_TCP -> 6
// eg. 0 = ip, 1 = icmp, 3 = ggp, 6 = tcp 8 = egp, 12 = pup, 17 = udp, 20 = hmp, 22 = xns-idp, 27 = rdp, 66 = rvd
    if (gettype($sd_ipc['socket_protocol']) == 'string')
    {
      if (($_def = getprotobyname($sd_ipc['socket_protocol'])) == -1)
        $sd_ipc['socket_protocol'] = SONIC_SOCK_PROTOCOL;
      else
        $sd_ipc['socket_protocol'] = $_def;

      unset($_def);
    }
    else
    {
      if (getprotobynumber($sd_ipc['socket_protocol']));
      else
        $sd_ipc['socket_protocol'] = SONIC_SOCK_PROTOCOL;
    }

    if ($sd_ipc['socket'])
    {
      @socket_close($sd_ipc['socket']);
//      usleep(100); //much?
    }

    if (($sd_ipc['socket'] = @socket_create($sd_ipc['socket_domain'], $sd_ipc['socket_type'], $sd_ipc['socket_protocol'])) == false && $fatal)
    {
      die('! ['.$_sd_pid.'] '.sonic_datef()
         .' [SONICD_IPC] Error: '.$sd_ipc['ip'].':'.$sd_ipc['port'].', socket_create(): '
         .socket_strerror(socket_last_error())
         ."\n"
         );
    }
    else if ($sd_ipc['socket'] == false)
    {
      if ($quiet);
      else
        echo '! ['.$_sd_pid.'] '.sonic_datef()
            .' [SONICD_IPC] Error: '.$sd_ipc['ip'].':'.$sd_ipc['port'].', socket_create(): '
            .socket_strerror(socket_last_error())
           ."\n";

      return false;
    }

    if ($sd_ipc['socket_options'])
      sonic_socket_set_options($sd_ipc['socket'], SOL_SOCKET, $sd_ipc['socket_options']);

    if (($_sbind = @socket_bind($sd_ipc['socket'], $sd_ipc['ip'], $sd_ipc['port'])) == false && $fatal)
    {
      die('! ['.$_sd_pid.'] '.sonic_datef()
         .' [SONICD_IPC] Error: '.$sd_ipc['ip'].':'.$sd_ipc['port'].', socket_bind(): '
         .socket_strerror(socket_last_error($sd_ipc['socket']))
         ."\n"
         );
    }
    else if ($_sbind == false)
    {
      if ($quiet);
      else
        echo '! ['.$_sd_pid.'] '.sonic_datef()
            .' [SONICD_IPC] Error: '.$sd_ipc['ip'].':'.$sd_ipc['port'].', socket_bind(): '
            .socket_strerror(socket_last_error($sd_ipc['socket']))
            ."\n";

      return false;
    }
    unset($_sbind);

    if ($sd_ipc['socket_type'] == SOCK_STREAM)
    {
      if (($_slisten = @socket_listen($sd_ipc['socket'], 5)) == false) // && $fatal)
      {
        die('! ['.$_sd_pid.'] '.sonic_datef()
           .' [SONICD_IPC] Error: '.$sd_ipc['ip'].':'.$sd_ipc['port'].', socket_listen(): '
           .socket_strerror(socket_last_error($sd_ipc['socket']))
           ."\n" 
           );
      }
      else if ($_slisten == false)
      {
        if ($quiet);
        else
          echo '! ['.$_sd_pid.'] '.sonic_datef()
              .' [SONICD_IPC] Error: '.$sd_ipc['ip'].':'.$sd_ipc['port'].', socket_listen(): '
              .socket_strerror(socket_last_error($sd_ipc['socket']))
              ."\n";

        return false;
      }
      unset($_slisten);
    }
  }

  if ($quiet);
  else
    echo '* IPC enabled'."\n".'- IPC socket info. -> '.$sd_ipc['ip'].':'.$sd_ipc['port']."\n";

  return true;
}

function sonic_daemon_IPC_socket_shutdown(&$sd_ipc, $opts = array('quiet' => true))
{
  $quiet = $opts['quiet'];

  if ($sd_ipc['socket'])
  {

    if ($quiet);
    else
    {
//      $pcntl_loaded = (extension_loaded('pcntl') && extension_loaded('posix'));
      $pcntl_loaded = SONICD_PCNTL;
      $_sd_pid = $pcntl_loaded ? posix_getpid() : getmypid();
      echo '= ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] Exit'."\n";
    }

    //&=, && 
    $ss_ret = @socket_getsockname($sd_ipc['socket'], $ssaddr, $ssport);
    $ss_ret = (@socket_shutdown($sd_ipc['socket'], 2) && $ss_ret);
    $ss_ret = (@socket_close($sd_ipc['socket']) && $ss_ret);

    if ($sd_ipc['socket_domain'] == AF_UNIX)
      $ss_ret = (@unlink($ssaddr) && $ss_ret);

    return $ss_ret;
  }
  
  return true;
}

function _sonic_daemon_IPC_session_shmsem_register($key)
{
  $slp = true;
  $session_keys = sonic_daemon_IPC_session($slp, '_sonic_daemon_IPC_session_shmsem_register', NULL, false);

  if (gettype($session_keys) == 'array');
  else
    $session_keys = array();

  $session_keys[$key] = true;

  return sonic_daemon_IPC_session($slp, '_sonic_daemon_IPC_session_shmsem_register', $session_keys, false);
}

function _sonic_daemon_IPC_session_shmsem_unregister($key)
{
  $slp = true;
  $session_keys = sonic_daemon_IPC_session($slp, '_sonic_daemon_IPC_session_shmsem_register', NULL, false);

  if (gettype($session_keys) == 'array');
  else
    $session_keys = array();

  if ($session_keys[$key])
  {
    sonic_daemon_IPC_session_shmsem_shutdown($key);
    unset($session_keys[$key]);
  }

  return sonic_daemon_IPC_session($slp, '_sonic_daemon_IPC_session_shmsem_register', $session_keys, false);
}

function _sonic_daemon_IPC_session_shmsem_unregister_all()
{
  $slp = true;
  $session_keys = sonic_daemon_IPC_session($slp, '_sonic_daemon_IPC_session_shmsem_register', NULL, false);

  if (gettype($session_keys) == 'array');
  else
    $session_keys = array();

  foreach ($session_keys as $key => $_kb)
  {
    sonic_daemon_IPC_session_shmsem_shutdown($key);
    unset($session_keys[$key]);
  }

  return sonic_daemon_IPC_session($slp, '_sonic_daemon_IPC_session_shmsem_register', $session_keys, false);
}

function sonic_daemon_IPC_session_shmsem_shutdown($key)
{
  if (($mem_key = crc32('sonic_daemon_IPC_session_memkey_shift'.$key)) > 0);
  else
    $mem_key *= -1;

  $mem_sem_key = &$mem_key;
  $mem_shm_key = &$mem_key;

  //semapores = access locking
  if ($mem_sem_id = sem_get($mem_sem_key, 1));
  else
  {
//    die('! Warning: main() failed: reason: unable to get semaphore'."\n");
//    return -1;
    return false;
  }

  if (sem_acquire($mem_sem_id));
  else
  {
    sem_release($mem_sem_id);

//    die('! Warning: main() failed: reason: unable to get semaphore '.$sem_id."\n");
//    return -2;
    return false;
  }

//attach 
  //actual shared memory
  if (($mem_shm_id = shm_attach($mem_shm_key, sonic_shm_memsize(PHP_INT_MAX, true))) == false)
  {
    sem_release($mem_sem_id);

//    die('! Warning: main() failed: reason: unable to attach shared memory'."\n");
//    return -3;
    return false;
  }

  //using index 1 for mem_info
  //no real way to tell if return value of false is because of error
  $memsize = @shm_get_var($mem_shm_id, 1); //muted as init call DNE

  $memsize += 0;

//LOCKPASS
  if ($memsize > 0)
  {
    if (($key_shmsem = crc32($key)) > 0);
    else
      $key_shmsem *= -1;
    
    $key_sem = &$key_shmsem;
    $key_shm = &$key_shmsem;

    //semapores = access locking
    if ($sem_id = sem_get($key_sem, 1));
    else
    {
      shm_detach($mem_shm_id);
      sem_release($mem_sem_id);

//      die('! Warning: main() failed: reason: unable to get semaphore'."\n");
//      return -4;
      return false;
    }

    if (sem_acquire($sem_id));
    else
    {
      shm_detach($mem_shm_id);
      sem_release($mem_sem_id);

      sem_release($sem_id);
 
//      die('! Warning: main() failed: reason: unable to get semaphore '.$sem_id."\n");
//      return -5;
      return false;
    }

//GET SHM SETUP
//attach
    //actual shared memory
    if (($shm_id = shm_attach($key_shm, $memsize)) == false)
    {
      shm_detach($mem_shm_id);
      sem_release($mem_sem_id);

      sem_release($sem_id);

//      die('! Warning: main() failed: reason: unable to attach shared memory'."\n");
//      return -6;
      return false;
    }

    sem_release($sem_id);
    shm_remove($shm_id);
    sem_remove($sem_id);
  }

  sem_release($mem_sem_id);
  shm_remove($mem_shm_id);
  sem_remove($mem_sem_id);

  return true;
}

//NB, apps that use this in threaded mode should really implement locking 
function sonic_daemon_IPC_session(&$lock_pass, $key, $session_var = null, $not_only_local = true)
{
  global	$_sd_ipc_service_info;

  global	$_sd_ipc_sessions;

  global	$_sd_ipc_sessions_shmsem_info;

  global	$_sd_ipc_sessions_lock_cache;

//debug
//    static $_entry_count = 0;
//    $_entry_count++;

  $_sisi = &$_sd_ipc_service_info;

  if (gettype($_sd_ipc_sessions) == 'array');
  else
    $_sd_ipc_sessions = array();

  if (gettype($lock_pass) == 'string');
  else if (gettype($lock_pass) == 'boolean')
  {
    if ($lock_pass)
    {
      mt_srand(sonic_srand_seeder());
      $lock_pass = md5(mt_rand());
    }
  }
  else
    $lock_pass = false;

  if ($key);
  else
    return false;

  //TODO: shared memory option
  //	  enhance loop for sleep timeouts
  if (true && $not_only_local)
  {
    if (($mem_key = crc32('sonic_daemon_IPC_session_memkey_shift'.$key)) > 0);
    else
      $mem_key *= -1;

    $mem_sem_key = &$mem_key;
    $mem_shm_key = &$mem_key;

    //semapores = access locking
    if ($mem_sem_id = sem_get($mem_sem_key, 1));
    else
    {
//      die('! Warning: main() failed: reason: unable to get semaphore'."\n");
//      return -1;
      return false;
    }

    if (sem_acquire($mem_sem_id));
    else
    {
      sem_release($mem_sem_id);

//      die('! Warning: main() failed: reason: unable to get semaphore '.$sem_id."\n");
//      return -2;
      return false;
    }

//attach 
    //actual shared memory
    if (($mem_shm_id = shm_attach($mem_shm_key, sonic_shm_memsize(PHP_INT_MAX, true))) == false)
    {
      sem_release($mem_sem_id);

//      die('! Warning: main() failed: reason: unable to attach shared memory'."\n");
//      return -3;
      return false;
    }

    //using index 1 for mem_info
    //no real way to tell if return value of false is because of error
    $memsize = @shm_get_var($mem_shm_id, 1); //muted as init call DNE

    $memsize += 0;

    //NB, only local process var
    $shmsem_info = &$_sd_ipc_sessions_shmsem_info;

    if ($shmsem_info[$key]['sem_id']);
    else
    {
      if (($shmsem_info[$key]['sem_id'] = crc32($key)) > 0);
      else
        $shmsem_info[$key]['sem_id'] *= -1;
    }
    
    if ($shmsem_info[$key]['shm_id']);
    else
    { 
      if (($shmsem_info[$key]['shm_id'] = crc32($key)) > 0);
      else
        $shmsem_info[$key]['shm_id'] *= -1;
    }

    $key_sem = &$shmsem_info[$key]['sem_id'];
    $key_shm = &$shmsem_info[$key]['shm_id'];

    //semapores = access locking
    if ($sem_id = sem_get($key_sem, 1));
    else
    {
      shm_detach($mem_shm_id);
      sem_release($mem_sem_id);

//      die('! Warning: main() failed: reason: unable to get semaphore'."\n");
//      return -4;
      return false;
    }

    if (sem_acquire($sem_id));
    else
    {
      shm_detach($mem_shm_id);
      sem_release($mem_sem_id);

      sem_release($sem_id);

//      die('! Warning: main() failed: reason: unable to get semaphore '.$sem_id."\n");
//      return -5;
      return false;
    }

//LOCKPASS
    if ($lock_pass && $memsize > 0)
    {
//GET SHM SETUP
//attach
      //actual shared memory
      if (($shm_id = shm_attach($key_shm, $memsize)) == false)
      {
        shm_detach($mem_shm_id);
        sem_release($mem_sem_id);

        sem_release($sem_id);

//        die('! Warning: main() failed: reason: unable to attach shared memory'."\n");
//        return -6;
        return false;
      }

//LOCK CACHE
//fetch: lock_cache
      //using index 2 for lock_cache
      //no real way to tell if return value of false is because of error
      $lock_cache_pass = shm_get_var($shm_id, 2);

//NB, depends on GET or SET see below
//      shm_detach($shm_id);
//      shm_remove($shm_id);
    }
    else if ($lock_pass)
    {
      $lock_cache_pass = NULL;
    }

//SET
    if ($key && isset($session_var))
    {
      if ($lock_cache_pass == $lock_pass)
      {
        $lock_cache_pass = $lock_pass;
      }
      else if ($lock_cache_pass)
      {
        shm_detach($mem_shm_id);
        sem_release($mem_sem_id);

        sem_release($sem_id);

//TODO: tweak time
        usleep(500);

//TODO: break the vicous cycle if next/leader of queue is zombified
//var_dump('SET:', $_entry_count);

        return sonic_daemon_IPC_session($lock_pass, $key, $session_var, false);
      }

      if ($lock_pass && $memsize > 0)
        //from lock_cache_pass init above
        shm_remove($shm_id);

//SET SHM SETUP
//      if ($lock_pass)
//        $lock_cache_pass = $lock_pass;

      $memsize = sonic_shm_memsize($lock_cache_pass, true) + sonic_shm_memsize($session_var, true);

      //from lock_cache_pass init above
      shm_remove($shm_id);

//attach:
      if (($shm_id = shm_attach($key_shm, $memsize)) == false)
      {
        shm_detach($mem_shm_id);
        sem_release($mem_sem_id);

        sem_release($sem_id);

//        die('! Warning: main() failed: reason: unable to attach shared memory'."\n");
//        return -7;
        return false;
      }

//SESSION VAR
      if ($lock_pass)
      {
//        $lock_cache_pass = $lock_pass;

//store: lock_cache_pass
        //using index 2 for lock_cache_pass
        if (shm_put_var($shm_id, 2, $lock_cache_pass));
        else
        {
          shm_detach($mem_shm_id);
          sem_release($mem_sem_id);

          sem_release($sem_id);
          shm_detach($shm_id);

//          die('! Warning: main() failed: reason: unable to put var2 into memory'."\n");
//          return -8;
          return false;
        }
      }

//store: session_var
      //using index 1, seeing we are tracking 1 memsize
      if (shm_put_var($shm_id, 1, $session_var));
      else
      {
        shm_detach($mem_shm_id);
        sem_release($mem_sem_id);

        sem_release($sem_id);
        shm_detach($shm_id);

//        die('! Warning: main() failed: reason: unable to put var1 into memory'."\n");
//        return -9;
        return false;
      }

      if ($lock_pass)
      {
        $lock_cache_pass = NULL;

//store: lock_cache_pass
        //using index 2 for lock_cache_pass
        if (shm_put_var($shm_id, 2, $lock_cache_pass));
        else
        {
          shm_detach($mem_shm_id);
          sem_release($mem_sem_id);

          sem_release($sem_id);
          shm_detach($shm_id);

//          die('! Warning: main() failed: reason: unable to put var2 into memory'."\n");
//          return -10;
          return false;
        }
      }

//TODO: store memsize
      //using index 1 for mem_info
      //no real way to tell if return value of false is because of error
      if (shm_put_var($mem_shm_id, 1, $memsize));
      else
      {
        shm_detach($mem_shm_id);
        sem_release($mem_sem_id);

        sem_release($sem_id);
        shm_detach($shm_id);

//        die('! Warning: main() failed: reason: unable to put var memsize into memory'."\n");
//        return -11;
        return false;
      }

      shm_detach($mem_shm_id);
      sem_release($mem_sem_id);

      shm_detach($shm_id);
      sem_release($sem_id);

//      shm_remove($shm_id);
//      sem_remove($sem_id);

//      $_sd_ipc_sessions[$key] = $session_var;

      return true;
    }

//GET
    if ($memsize > 0)
    {
      if ($lock_pass)
        //from lock_cache_pass init above
        shm_detach($shm_id);
    }
    else
    {
      if ($lock_pass)
      {
//SET SHM SETUP
        $lock_cache_pass = $lock_pass;

        $memsize = sonic_shm_memsize($lock_cache_pass, true);

        //actual shared memory
        if (($shm_id = shm_attach($key_shm, $memsize)) == false)
        {
          shm_detach($mem_shm_id);
          sem_release($mem_sem_id);

          sem_release($sem_id);

//          die('! Warning: main() failed: reason: unable to attach shared memory'."\n");
//          return -12;
          return false;
        }

//store: lock_cache_pass
        //using index 2 for lock_cache_pass
        if (shm_put_var($shm_id, 2, $lock_cache_pass));
        else
        {
          shm_detach($mem_shm_id);
          sem_release($mem_sem_id);

          shm_detach($shm_id);
          sem_release($sem_id);

//          die('! Warning: main() failed: reason: unable to put var2 into memory'."\n");
//          return -13;
          return false;
        }

//TODO: store memsize
        //using index 1 for mem_info
        //no real way to tell if return value of false is because of error
        if (shm_put_var($mem_shm_id, 1, $memsize));
        else
        { 
          shm_detach($mem_shm_id);
          sem_release($mem_sem_id);

          sem_release($sem_id);
          shm_detach($shm_id);

//          die('! Warning: main() failed: reason: unable to put var memsize into memory'."\n");
//          return -14;
          return false;
        }
      
      }

      shm_detach($mem_shm_id);
      sem_release($mem_sem_id);

      shm_detach($shm_id);
      sem_release($sem_id);

      return NULL;
    }

    shm_detach($mem_shm_id);
    sem_release($mem_sem_id);

    if ($lock_cache_pass == $lock_pass) {}
    else if ($lock_cache_pass)
    {
      sem_release($sem_id);

//TODO: tweak time
      usleep(500);

//TODO: break the vicous cycle if next/leader of queue is zombified
//var_dump('SET:', $_entry_count);

      return sonic_daemon_IPC_session($lock_pass, $key, $session_var, false);
    }

//GET SHM SETUP
    //actual shared memory
    if (($shm_id = shm_attach($key_shm, $memsize)) == false)
    {
      sem_release($sem_id);

//      die('! Warning: main() failed: reason: unable to attach shared memory'."\n");
//      return -15;
      return false;
    }

//SESSION_VAR
    if ($lock_pass)
    {
      $lock_cache_pass = $lock_pass;

//store: lock_cache_pass
      //using index 2 for lock_cache_pass
      if (shm_put_var($shm_id, 2, $lock_cache_pass));
      else
      {
        sem_release($sem_id);
        shm_detach($shm_id);

//        die('! Warning: main() failed: reason: unable to put var2 into memory'."\n");
//        return -16;
        return false;
      }
    }

    //only using index 1
    //no real way to tell if return value of false is because of error
    $session_var = shm_get_var($shm_id, 1);

    shm_detach($shm_id);
    sem_release($sem_id);

//    return $_sd_ipc_sessions[$key];
    return $session_var;
  }
  else if (gettype($_sisi['handlers_array']) == 'array' && in_array('http', $_sisi['handlers_array']) && $not_only_local)
  {
    if ($host = &$_sisi['ip']);
    else
//      die('! Warning: main() failed: reason: destination host not defined in config'."\n");
      return false;

    if (isset($_sisi['port']))
    {
      $port = &$_sisi['port'];
      $port += 0;
    }
    else
//      die('! Warning: main() failed: reason: destination port not defined in config'."\n");
      return false;

    //SETUP CONNECTION OUT AND BIND SRC IP
    //helper suff. (setup for inbound socket type stream)
    if (isset($_sisi['socket_domain']))
    {
      $s_domain = &$_sisi['socket_domain'];

      if ($_def = @constant($s_domain))
        $s_domain = $_def;

      $s_domain += 0;
    }
    else
      $s_domain = SONIC_SOCK_DOMAIN; //eg. AF_INET

    if (isset($_sisi['socket_type']))
    {
      $s_type = &$_sisi['socket_type'];

      if ($_def = @constant($s_type))
        $s_type = $_def;

      $s_type += 0;
    }
    else
      $s_type = SONIC_SOCK_TYPE; //eg. SOCK_STREAM

    if (isset($_sisi['socket_protocol']))
    {
      $s_protocol = &$_sisi['socket_protocol'];

      if ($_def = @constant($s_protocol))
        $s_protocol = $_def;

      $s_protocol += 0;
    }
    else
      $s_protocol = SONIC_SOCK_PROTOCOL; //eg. SOL_TCP

    unset($_def);

    //string tcp -> SOL_TCP map check
    if (gettype($s_protocol) == 'string')
    {
      if (($_sp = getprotobyname($s_protocol)) == -1)
        $s_protocol = SONIC_SOCK_PROTOCOL;
      else
      {
        $s_protocol = $_sp;
        unset($_sp);
      }
    }

    if (($s_options = &$_sisi['socket_options']) != NULL);
    else
      $s_options = SONIC_SOCK_OPTIONS;

    if (($sock_out = @socket_create($s_domain, $s_type, $s_protocol)) == false)
//      die('! Warning: socket_create() failed: reason: '.socket_strerror(socket_last_error())."\n");
      return false;

//TODO: rehash option set
//    if ($s_options)
      sonic_socket_set_options($sock_out, SOL_SOCKET, $s_options);

    if ($src_ip)
      if (@socket_bind($sock_out, $src_ip) == false)
//        die('! Warning: socket_bind() failed: reason: '.socket_strerror(socket_last_error($sock_out))."\n");
//        return false;
        return -2;

    if ($s_type == SOCK_STREAM)
    {
      if (@socket_connect($sock_out, $host, $port) == false)
        die('! Warning: socket_connect() failed: reason: '.socket_strerror(socket_last_error($sock_out))."\n");
//        return false;
//        return -1;
    }

    if (isset($session_var))
    {
      $type = gettype($session_var);
      switch ($type)
      {
        case 'boolean':
        case 'integer':
        case 'NULL':
          $svhr = &$session_var;
        break;

        case 'array':
          $svhr = sonic_array2str($session_var);
        break;

        default:
          $svhr = var_export($session_var, true);
      }

      $sihr = 'GET /?op=IPC-session&name=set&key='.$key
             .'&type='.$type
             .'&var='.base64_encode($svhr)
             .($lock_pass ? '&lock_pass='.$lock_pass : '')
             .' HTTP/1.0'
             ."\r\n\r\n";
    }
    else
    {
      $sihr = 'GET /?op=IPC-session&name=get&key='.$key.($lock_pass ? '&lock_pass='.$lock_pass : '').' HTTP/1.0'
             ."\r\n\r\n";
    }

    if ($s_type == SOCK_STREAM)
      $i_bytes = @socket_write($sock_out, $sihr, strlen($sihr));
    else
      $i_bytes = @socket_sendto($sock_out, $sihr, strlen($sihr), 0, $host, $port);

    if ($i_bytes < 0 OR $i_bytes == false)
//      echo '! main() echo failed: reason: '.socket_strerror(socket_last_error($sock_out_mux))."\n";
//      return false;
      return -7;

    socket_shutdown($sock_out, 1); //shutdown write
    usleep(50);

    $sihres = '';

    do
    {
      $ss_read = array($sock_out);
      $ss_write = NULL;
      $ss_except = NULL;
      $ss_changed_sockets = socket_select($ss_read, $ss_write, $ss_except, 0, 1);

      if ($ss_changed_sockets > 0)
      {
        if (false == ($buf_out = socket_read($sock_out, SONIC_IPC_BYTE_SIZE, PHP_BINARY_READ)))
          break;

        $sihres .= $buf_out;
      }
    }
    while (true); unset($buf_out);

    socket_shutdown($sock_out, 0);//close reading
    socket_close($sock_out); //finaly we can free resource

    list($sihres_head, $sihres_body) = preg_split('/\r{0,1}\n\r{0,1}\n/', $sihres, 2);

    $sihres_h = sonic_header2array($sihres_head);

    if (isset($session_var))
    {
      if ($sihres_h['Sonic-IPC-Lock-Failure'])
      {
//TODO: tweak time
        usleep(500);

//TODO: break the vicous cycle if next/leader of queue is zombified
//$out_headers['Sonic-IPC-Lock-Pass'] //use force?
//var_dump('SET:', $_entry_count);

        return sonic_daemon_IPC_session($lock_pass, $key, $session_var, true);
      }

      return $sihres_body == 'OK';
    }
    else
    {
//      $sihres_h = sonic_header2array($sihres_head);

      if ($sihres_h['Sonic-IPC-Lock-Failure'])
      {
//TODO: tweak time
        usleep(500);

//TODO: break the vicous cycle if next/leader of queue is zombified
//$out_headers['Sonic-IPC-Lock-Pass'] //use force?
//var_dump('GET:', $_entry_count);

        return sonic_daemon_IPC_session($lock_pass, $key, NULL, true);
      }

      if ($var_type = $sihres_h['Sonic-IPC-Variable-Type']);
      else
      {
        $var_type = 'string';
        $var = var_export($sihres_body, true);
      }

      switch ($var_type)
      {
        case 'array':
          $var = sonic_str2array($sihres_body);
        break;

        case 'boolean':
        case 'integer':
        case 'NULL':
          $var = settype($sihres_body, $var_type);
        break;
      }

      return $var;
    }
  }
  //local process only
  //NB, will return false instead of usleep repeat as this is a last step process
  else
  {
    $lock_cache = &$_sd_ipc_sessions_lock_cache;
//SET

    if ($key && isset($session_var))
    {
      if ($lock_cache[$key]['locked'] == $lock_pass) {}
      else if ($lock_cache[$key]['locked'])
      {
        $lock_pass = $lock_cache[$key]['locked']; //incase force is needed
        return false;
      }

      if ($lock_pass)
      {
        $lock_cache[$key]['locked'] = $lock_pass;
        $_sd_ipc_sessions[$key] = $session_var;
        unset($lock_cache[$key]['locked']);
      }
      else
        $_sd_ipc_sessions[$key] = $session_var;

      return true;
    }
//GET

    if ($lock_cache[$key]['locked'] == $lock_pass) {} 
    else if ($lock_cache[$key]['locked'])
    {
      //getters might inot know original lock_pass so no point returning a different one
      //plus getters can wait
      $lock_pass = false;
      //if they are getting know where and want to force set later
      //via returned stale lock pass (better than null)
      return $_sd_ipc_sessions[$key];
    }

    if ($lock_pass)
      $lock_cache[$key]['locked'] = $lock_pass;

    return $_sd_ipc_sessions[$key];
  }
}

//nb, always returns array
function sonic_daemon_load_config_services($config = array(), $opts = array('fatal' => false, 'quiet' => true))
{
//  $pcntl_loaded = (extension_loaded('pcntl') && extension_loaded('posix'));
  $pcntl_loaded = SONICD_PCNTL;
  $_sd_pid = $pcntl_loaded ? posix_getpid() : getmypid();

  $fatal = $opts['fatal'];
  $quiet = $opts['quiet'];
  $prelog = $opts['prelog'];

  if ($config['services'])
  {
    $config_services = preg_split('/\s*,\s*/', $config['services']);
  }
  else if ($fatal)
  {
    die('! ['.$_sd_pid.'] '.sonic_datef().' [SONICD_SERV] Error: no services are set to load'."\n");
  }
  else
    return array(array(), 0);

  $services = array();

  //load stream parent config
  while (list($i, $service) = each($config_services))
  {
    $services[$service]['stream'] = &$config['service-'.$service.'-stream'];

    //misc: load stream
    $stream_name = $services[$service]['stream'];
    $stream = $_SERVER[SONIC_LIB].DIRECTORY_SEPARATOR.'streams'.DIRECTORY_SEPARATOR.'sonic.stream.'.$stream_name.'.php';
    if (file_exists($stream))
    {
      require_once($stream);
    }
    else 
    {
      unset($services[$service]);

      echo '! ['.$_sd_pid.'] '.sonic_datef()
          .' [SONICD_SERV] Error: can not load '.($stream_name ? $stream_name.' ' : '')
          .'stream for '.$service.' service'."\n";

      if ($fatal)
      {
        if ($_sd_config['output'] == 'buffer')
        {
          //fork buff safety clear
          $out = ob_get_contents();
          ob_end_clean();

          if (!$_sd_config['buffer_quiet'])
            echo $out;

          if ($_sd_config['log'] && $out != '')
            sonic_file_mode_contents($_sd_config['log'], $out, 'a');

          unset($out);
        }

        exit();
      }

      continue;
    }

    if (!$quiet)
    {
      if ($prelog)
        echo '+ ['.$_sd_pid.'] '.sonic_datef().' [SONICD_SERV] ';
      else
        echo '+ ';

      echo 'reading service information for ['.$service.']'."\n";
    }

    //helper setup for stream
    $services[$service]['name'] = $service;

    if ($services[$service]['max'] = &$config['service-'.$service.'-max'])
      $services[$service]['max'] += 0;
    else
      $services[$service]['max'] = 1;

    $threads_total += $services[$service]['max']; 

    if ($services[$service]['run_once'] = &$config['service-'.$service.'-run_once'])
      $services[$service]['run_once'] = true;
    else
      $services[$service]['run_once'] = false;

    //NB, NO DEFAULTs (not CORE-essential, but important)
    //daemon will twiddle based on NULL after stream constructor setup
    //NB, verbose is twiddled by stream
    if ($_thread_out = &$config['service-'.$service.'-thread_out'])
    {
      $_thread_out = strtolower($_thread_out);
      switch ($_thread_out)
      {
        case 'flush':
        case 'implicit':
        case 'default':
        case 'normal':
        case 'standard':
          $services[$service]['thread_out'] = 'flush';
        break;

        case 'buffer':
        case 'buff':
        case 'log':
        case 'cache':
          $services[$service]['thread_out'] = 'buffer';
        break;
      }
    }
//    else //leave NULL, daemon will set later
//      $services[$service]['thread_out'] = NULL;

    //leave NULL, daemon will set later
    if ($config['service-'.$service.'-thread_buffer_quiet'] == NULL);
    else
      $services[$service]['thread_buffer_quiet'] = true;

    //NB, log sanity check done by stream constructor
    if ($config['service-'.$service.'-thread_log'] == NULL);
    else
    {
      $services[$service]['thread_log'] = &$config['service-'.$service.'-thread_log'];
    }

    if (isset($config['service-'.$service.'-port']) OR ($services[$service]['socket_domain'] == AF_UNIX))
    {
      $services[$service]['port'] = &$config['service-'.$service.'-port'];
      $services[$service]['port'] += 0;

      //helper suff. (setup for inbound socket type stream)
      if (isset($config['service-'.$service.'-socket_domain']))
      {
        $services[$service]['socket_domain'] = &$config['service-'.$service.'-socket_domain'];

        if ($_def = @constant($services[$service]['socket_domain']))
          $services[$service]['socket_domain'] = $_def;

        $services[$service]['socket_domain'] += 0;
      }
      else
        $services[$service]['socket_domain'] = SONIC_SOCK_DOMAIN;

      if (isset($config['service-'.$service.'-socket_type']))
      {
        $services[$service]['socket_type'] = &$config['service-'.$service.'-socket_type'];

        if ($_def = @constant($services[$service]['socket_type']))
          $services[$service]['socket_type'] = $_def;

        $services[$service]['socket_type'] += 0;
      }
      else
        $services[$service]['socket_type'] = SONIC_SOCK_TYPE;

      if (isset($config['service-'.$service.'-socket_protocol']))
      {
        $services[$service]['socket_protocol'] = &$config['service-'.$service.'-socket_protocol'];

        if ($_def = @constant($services[$service]['socket_protocol']))
          $services[$service]['socket_protocol'] = $_def;

        $services[$service]['socket_protocol'] += 0;
      }
      else
        $services[$service]['socket_protocol'] = SONIC_SOCK_PROTOCOL;

      if (($services[$service]['socket_options'] = &$config['service-'.$service.'-socket_options']) != NULL);
      else
        $services[$service]['socket_options'] = SONIC_SOCK_OPTIONS;

      //help ness.
      if ($services[$service]['ip'] = &$config['service-'.$service.'-ip']);
      else
      {
        switch ($services[$service]['socket_domain'])
        {
          case AF_INET:
            $services[$service]['ip'] = '0.0.0.0';
          break;
          case AF_INET6:
            $services[$service]['ip'] = 'ff:ff:ff:ff:ff:ff';
          break;
          case AF_UNIX: //port = 0
            $services[$service]['ip'] = SONIC_TMP_PATH.DIRECTORY_SEPARATOR.'sonic_'.md5(time()).'.sock';
            $services[$service]['port'] = 0;
          break;
        }
      }
    }
    else unset($services[$service]['port']);

  }
  reset($config_services);

  //load stream child config
  while (list($i, $service) = each($config_services))
  {
    while (list($config_key, $config_value) = each($config))
    {
      $config_key_match = array();
      //byte_size really lives in stream_config space putting here to force compliance
      if (preg_match('/service-'.$service.'-((?!stream|max|ip|port|byte_size|socket_domain|socket_type|socket_protocol|socket_options|thread_out|thread_buffer_quiet|thread_log|run_once).*)$/i', $config_key, $config_key_match))
      {
        if (preg_match('/^[0-9]+$/', $config_value))
          $config_value += 0;

        $services[$service]['stream_config'][strtolower($config_key_match[1])] = $config_value;
      }
      else
        continue;
    }
    reset($config);
  }
  reset($config_services);

  return array($services, $threads_total);
}

function sonic_daemon_load_services(&$services, $opts = array('fatal' => false, 'quiet' => true))
{
//  $pcntl_loaded = (extension_loaded('pcntl') && extension_loaded('posix'));
  $pcntl_loaded = SONICD_PCNTL;
  $_sd_pid = $pcntl_loaded ? posix_getpid() : getmypid();

  $fatal = $opts['fatal'];
  $quiet = $opts['quiet'];

  $fatal_ok = true;

  if (gettype($services) == 'array');
  else
  {
    echo '! ['.$_sd_pid.'] '.sonic_datef().' [SONICD_SERV] Error: '
	.'sonic_daemon_load_services(): services variable passed is not array'
	."\n";

    if ($fatal)
    {
      if ($_sd_config['output'] == 'buffer')
      {
        //fork buff safety clear
        $out = ob_get_contents();
        ob_end_clean();

        if (!$_sd_config['buffer_quiet'])
          echo $out;

        if ($_sd_config['log'] && $out != '')
          sonic_file_mode_contents($_sd_config['log'], $out, 'a');

        unset($out);
      }

      exit();
    }
    else
      return false;
  }

  while (list($service, $service_config) = each($services))
  {
    //unique key which _may_ used by service for locking purposes
    mt_srand(sonic_srand_seeder());
    $services[$service]['key'] = md5(mt_rand());

    if (isset($service_config['port'])) //socket setup trigger, check above
    {
//word map tcp -> SOL_TCP -> 6
// eg. 0 = ip, 1 = icmp, 3 = ggp, 6 = tcp 8 = egp, 12 = pup, 17 = udp, 20 = hmp, 22 = xns-idp, 27 = rdp, 66 = rvd
      if (gettype($service_config['socket_protocol']) == 'string')
      {
        if (($service_config['socket_protocol'] = getprotobyname($service_config['socket_protocol'])) == -1)
          $services[$service]['socket_protocol'] = SONIC_SOCK_PROTOCOL;
      }
      else
      {
        if (getprotobynumber($service_config['socket_protocol']));
        else
          $services[$service]['socket_protocol'] = SONIC_SOCK_PROTOCOL;
      }

      if ($services[$service]['socket'])
      {
        @socket_close($services[$service]['socket']);
//        usleep(100); //much?
      }

      if (($services[$service]['socket'] = @socket_create($services[$service]['socket_domain'], $service_config['socket_type'], $service_config['socket_protocol'])) == false && $fatal)
      {
        die('! ['.$_sd_pid.'] '.sonic_datef()
           .' [SONICD_SERV] Error: ['.$service.'] '.$service_config['ip'].':'.$service_config['port'].', socket_create(): '
           .socket_strerror(socket_last_error())."\n");
      }
      else if ($services[$service]['socket'] == false)
      {
        if ($quiet);
        else
          echo '! ['.$_sd_pid.'] '.sonic_datef()
              .' [SONICD_SERV] Error: ['.$service.'] '.$service_config['ip'].':'.$service_config['port'].', socket_create(): '
              .socket_strerror(socket_last_error())."\n";

        unset($services[$service]);
        $fatal_ok = false;
        continue;
//        return false;
      }

      if ($service_config['socket_options'])
        sonic_socket_set_options($services[$service]['socket'], SOL_SOCKET, $service_config['socket_options']);

      if (($_sbind = @socket_bind($services[$service]['socket'], $service_config['ip'], $service_config['port'])) == false && $fatal)
      {
        die('! ['.$_sd_pid.'] '.sonic_datef()
           .' [SONICD_SERV] Error: ['.$service.'] '.$service_config['ip'].':'.$service_config['port'].', socket_bind(): '
           .socket_strerror(socket_last_error($services[$service]['socket']))."\n"
           );
      }
      else if ($_sbind == false)
      {
        if ($quiet);
        else
          echo '! ['.$_sd_pid.'] '.sonic_datef()
              .' [SONICD_SERV] Error: ['.$service.'] '.$service_config['ip'].':'.$service_config['port'].', socket_bind(): '
              .socket_strerror(socket_last_error($services[$service]['socket']))."\n";

        unset($services[$service]);
        $fatal_ok = false;
        continue;
//        return false;
      }
      unset($_sbind);

      if ($service_config['socket_type'] == SOCK_STREAM)
      {
        if (($_slisten = @socket_listen($services[$service]['socket'], 5)) == false && $fatal)
        {
          die('! ['.$_sd_pid.'] '.sonic_datef()
             .' [SONICD_SERV] Error: ['.$service.'] '.$service_config['ip'].':'.$service_config['port'].', socket_listen(): '
             .socket_strerror(socket_last_error($services[$service]['socket']))."\n"
             );
        }
        else if ($_slisten == false)
        {
          if ($quiet);
          else
          {
            echo '! ['.$_sd_pid.'] '.sonic_datef()
                .' [SONICD_SERV] Error: ['.$service.'] '.$service_config['ip'].':'.$service_config['port'].', socket_listen(): '
                .socket_strerror(socket_last_error($services[$service]['socket']))."\n";
          }

          unset($services[$service]);
          $fatal_ok = false;
          continue;
//          return false;
        }
        unset($_slisten);
      }
    }
  }
  reset($services);

  return $fatal_ok;
//  return true;
}

//SPARKLED EYE
//TODO?: move rest of globals here also *shrug*
global	$_sd_ipc_sessions;
global	$_sd_ipc_sessions_lock_cache;
global	$_sd_ipc_sessions_shmsem_info;

//SELECT and setup THREAD or ROTATE mode
//AKA SONICD_THREADED in logs, $_sd_threaded in code
//if (extension_loaded('pcntl') && extension_loaded('posix'))
if (SONICD_PCNTL)
{
  global	$_sd_daemon_thread_handler_lock;

  $_sd_daemon_thread_handler_lock = false;

  function sonic_daemon_thread_kill($tpids = array(), $opts = array('quiet' => true, 'verbose' => 0))
  {
//    static	$entry_count = 0;

    global	$_sd_threads, $_sd_threads_shutdown;

//    $entry_count++;

    $quiet = $opts['quiet'];
    $verbose = $opts['verbose'];

    switch ($_sd_threads_shutdown)
    {
      case 0:
        $tsig = 'SIGINT';
      break;

      case 1:
        $tsig = 'SIGTERM';
      break;

      case 2:
        $tsig = 'SIGQUIT';
      break;

      default:
        $tsig = 'SIGKILL';
    }

    $pk_ret = true;
    foreach ($tpids as $tpid => $tpidi)
    {
      //only ours
      if (in_array($tpid, array_keys($_sd_threads)));
      else
        continue;

      if ($verbose > 1)
        echo '$ ['.posix_getpid().'] '.sonic_datef()
            .' [SONICD_THREADED] Notice: sending '.$tsig.' signal to ['.$tpid.']'
            ."\n";

//      $pk_ret &= posix_kill($tpid, constant($tsig)); //not sure if bitwise or logic assignment
      $pk_ret = (posix_kill($tpid, constant($tsig)) && $pk_ret);

      usleep(150000);
    }

    $ret_time_start = $time_start = time();
    $ret_time_limit = 30;
    $time_limit = 5; //seconds each thread
    $si_x = 1;

    foreach ($tpids as $_pid => $_info)
    {
      //only ours
      if (in_array($tpid, array_keys($_sd_threads)));
      else
        continue;

      //SIGSTOP or SIGKILL will cause non graceful shutdown
      $i = 0;
      $j = 1;
      $k = true;
      while ($k)
      {
        $i++;

        if (($ret_time_start + $ret_time_limit) > time())
        {
//          break;
          return $pk_ret;
        }

        if ($_sd_threads[$_pid])
        {
          if (($time_start + $time_limit) > time())
          {
//echo 'still waiting for ('.$_pid.')'."\n";
            usleep(150000);
          }
          else if ($j > $si_x)
          {
            if ($quiet);
            else
              echo '$ ['.posix_getpid().'] '.sonic_datef()
                  .' [SONICD_THREADED] Notice: main(): ['.$_pid.'] zombified by '.$si_x.' signals x time(s) wait of '.$time_limit.' seconds'."\n";

//echo 'issue kill or report zombie error'."\n";
//echo 'keep old service, threads will repopulate, or throw skip warning and remove service (zombies will remain regardless)'."\n";
            $_time_start = time();
            $_possible_recover = false;

            while (($_time_start + round($time_limit / 2)) > time())
            {
              $pk_ret = (posix_kill($tpid, SIGKILL) && $pk_ret);

              usleep(150000);

              if ($_sd_threads[$_pid])
                usleep(150000);
              else
                $_possible_recover = true;
            }
            unset($_time_start);

            if ($_possible_recover)
            {
              if ($quiet);
              else
                echo '! ['.posix_getpid().'] '.sonic_datef()
                    .' [SONICD_THREADED] Warning: main(): ['.$_pid.'] was non cleanly KILLED'."\n";

//            $k = true;
              break;
            }
            unset($_possible_recover);

            $k = false;
          }
          else
          {
            $j++;
            $time_start = time();
//echo 'sending int again to ('.$_pid.')'."\n";
            $pk_ret = (posix_kill($tpid, constant($tsig)) && $pk_ret);
//echo 'still waiting after retry for ('.$_pid.')'."\n";
            usleep(150000);
          }
        }
        else
          $k = false;
      }
//var_dump($i);
    }

    return $pk_ret;
  }

  function sonic_daemon_thread_info($opts = array('quiet' => true, 'verbose' => 0))
  {
    global	$_sd_threads, $_sd_threads_service, 
		$_sd_threads_service_active, 
		$_sd_threads_shutdown, 
		$_sd_daemon_thread_handler_lock;

    $fatal = $opts['fatal'];
    $quiet = $opts['quiet'];
    $verbose = $opts['verbose'];

    //zombie, aborted check
    $losties = array();
    $pid_izombie_pausenot_j = 0;
    foreach ($_sd_threads as $_sdt_pid => $_sdt_adt)
    {
      $pid_izombie_pausenot_i = 0;
      $pid_izombie_pausenot = false;
      if (sizeof($_sd_threads_service_active) > 0)
      {
        foreach ($_sd_threads_service_active as $_sdts => $_sdts_pids)
        {
//          $pid_izombie_pausenot |= in_array($_sdt_pid, array_keys($_sdts_pids));  //not sure if bitwise or logic assignment
          $pid_izombie_pausenot = (in_array($_sdt_pid, array_keys($_sdts_pids)) OR $pid_izombie_pausenot);
        }
      }
      else
        $pid_izombie_pausenot = false;
        
      if ($pid_izombie_pausenot);
      else
      {
        if ($verbose > 1)
          echo '$ ['.posix_getpid().'] '.sonic_datef()
                .' [SONICD_THREADED] Notice: zombie ['
                .$_sdt_adt['name'].', '.$_sdt_pid
                .'] still alive'
                ."\n";

/*
      while ($_sd_daemon_thread_handler_lock) { usleep(250); }

      $_sd_daemon_thread_handler_lock = true;
           
      if (gettype($_sd_threads_service[$_sdt_adt['name']]) == 'array');
      else
        $_sd_threads_service[$_sdt_adt['name']] = array();

      $_sd_threads_service[$_sdt_adt['name']][$_sdt_pid] = &$_sd_threads[$_sdt_pid];

      $_sd_daemon_thread_handler_lock = false;
*/
  
        $losties[$_sdt_pid] = $_sd_threads[$_sdt_pid]['name'];
        $pid_izombie_pausenot_i++;
      }
      $pid_izombie_pausenot_j += $pid_izombie_pausenot_i;
    }

    if ($pid_izombie_pausenot_j > 0)
    {
      if ($quiet);
      else
        echo '! ['.posix_getpid().'] '.sonic_datef()
//              .' [SONICD_THREADED] Warning: sonic_daemon_thread_info(): total zombie thread count is \''
              .' [SONICD_THREADED] Warning: total zombie thread count is \''
              .$pid_izombie_pausenot_j
              .'\''
              ."\n";
    }
//    unset($pid_izombie_pausenot, $pid_izombie_pausenot_i, $pid_izombie_pausenot_j);
    return $losties;
  }

  function sonic_daemon_thread_shutdown($sig = NULL)
  {
    global	$_sd_threads_shutdown;

    //NB mode=thread, child will enter here, will also have a _copy_
    //of threads array but only up until its spawn 

    static $_sd_tsc = 0;
    $_sd_tsc++;
    $_sd_threads_shutdown = $_sd_tsc;

/*
    //debug
    switch ($_sd_tsc)
    {
      case 1:
        $_sd_tsco = 'initiated';
      break;

      default:
        $_sd_tsco = 'attempt #'.$_sd_tsc;
    }

    echo '$ ['.posix_getpid().'] '.sonic_datef()
         .' [SONICD_THREADED] Notice: services SHUTDOWN '
	 .$_sd_tsco
	 ."\n";
*/
  }

  //forked/threaded child to local parent thread handler
  //AFAIK this function will stop the entire daemon
  //eg. no more death catching during current death catch
  //lock safey implemented anyway - chances are never used
  //
  function sonic_daemon_thread_handler($sig = NULL)
  {
    global	$_sd_threads, $_sd_threads_service, 
		$_sd_threads_service_active, 
		$_sd_threads_shutdown,
		$_sd_daemon_thread_handler_lock;

    if ($sig == SIGCHLD);
    else
      return;

    //parent pid of child pid that died
    $dead_ppid = pcntl_waitpid(-1, $status);

    //race helper between parent assign obj pid and child death 
    //balance here or in obj->thread_main()::run()
    usleep(1);

    //total # of threads
    if (array_key_exists($dead_ppid, $_sd_threads));
    else //not ours to manage
      return;

    while ($_sd_daemon_thread_handler_lock) { usleep(250); }
    $_sd_daemon_thread_handler_lock = true;

    //total # of threads of type, name stream service
    foreach ($_sd_threads_service as $live_service => $live_pids)
    {
      foreach ($live_pids as $live_pid => $live_pid_adt)
      {
        if ($live_pid == $dead_ppid)
        {
          unset($_sd_threads_service[$live_service][$live_pid]);

          if ($_sd_threads_service_active[$live_service][$live_pid])
            unset($_sd_threads_service_active[$live_service][$live_pid]);
        }
      }
    }

    $dead_name = $_sd_threads[$dead_ppid]['name'];
    $dead_quiet = $_sd_threads[$dead_ppid]['quiet'];
    $dead_verbose = $_sd_threads[$dead_ppid]['verbose'];

    if (in_array($dead_name, array_keys($_sd_threads_service_active)));
    else
    {
      if (sizeof($_sd_threads_service[$dead_name]) < 2) // == 1
      {
        if ($_sd_threads[$dead_ppid]['socket'])
        {
          if ($dead_verbose > 0)
            echo '! ['.$dead_ppid.'] '.sonic_datef()
                .' ['.$dead_name .'] Warning: sonic_daemon_thread_handler(): connection shutdown and close'
                ."\n";

//NB, socket var might not be valid resource at this point
          if ($_sd_threads[$dead_ppid]['socket_domain'] == AF_UNIX)
          {
            @socket_getsockname($_sd_threads[$dead_ppid]['socket'], $dead_addr, $dead_port);
            @unlink($dead_addr);
          }
          @socket_shutdown($_sd_threads[$dead_ppid]['socket'], 2);
          @socket_close($_sd_threads[$dead_ppid]['socket']);
        }
        unset($_sd_threads_service[$dead_name]);
      }
      //this (last) pid (of service) is late still having handler (with unprobable chance of entry)
      else if (sizeof($_sd_threads_service[$dead_name]) < 1) // == 0
      {
        if ($dead_verbose > 0)
          echo '! ['.$dead_ppid.'] '.sonic_datef()
              .' ['.$dead_name .'] Warning: sonic_daemon_thread_handler(): last round cleanup'
              ."\n";

        unset($_sd_threads_service[$dead_name]);
      }
    }
//witchhunt end

    unset($_sd_threads[$dead_ppid]);

    $_sd_daemon_thread_handler_lock = false;

    if ($dead_quiet);
    else
      echo '= ['.$dead_ppid.'] '.sonic_datef()
          .' ['.$dead_name .'] Exit'
          ."\n";

  }

  //required for sig/thread handler to work >= php v4.3.0+
  declare(ticks = 1);

  //set thread handler for sig types
  pcntl_signal(SIGINT,  'sonic_daemon_thread_shutdown');	//SIGINT=
  pcntl_signal(SIGCHLD, 'sonic_daemon_thread_handler');		//SIGCHLD=17

  //local manager vars				//THREAD INFO
  global	$_sd_threads, $_sd_threads_service, $_sd_threads_service_active, $_sd_threads_shutdown;

  $_sd_threads = array();			//main index
  $_sd_threads_service = array();		//main index (indexed by service)
  $_sd_threads_service_active = array();	//main index (indexed by service, current instance threads only)
  $_sd_threads_shutdown = 0;		//thread shutdown trigger and level of trigger

  $_sd_threaded = true;				//internal flag
  $_sd_threaded_pid = posix_getpid();
						//main index total thread count -> sizeof($_sd_threads);
  $_sd_threads_total = 0;			//secondary index total thread count (current instance threads only)
}
else
  $_sd_threaded = false;

//TODO: doc
if ($_sd_threaded)
  $_sd_pid = &$_sd_threaded_pid;
else
  $_sd_pid = getmypid();

include($_SERVER['SONIC_LIB'].'/sonic.inc');
include($_SERVER['SONIC_LIB'].'/sonic.stream.inc');

//TODO, actually use this apart from FATAL?
$_sd_exts = array('sockets');
$_sd_exts_ok = sonic_ext_check($_sd_exts);
//var_dump($_sd_exts, $_sd_exts_ok);
unset($_sd_exts, $_sd_exts_ok);

$SONIC_ARG = sonic_args($_SERVER['argv']);
//var_dump($_SERVER['argv'], $SONIC_ARG);

if ($SONIC_ARG['h'] OR $SONIC_ARG['help'])
{
  echo <<< EOF
Usage: [ --quiet | --help ]

        -i, --ini       SONIC configuration/settings file
                        (default: env SONIC_CONF/sonic.ini)

	-q, --quiet	Surpess headers
	-v, --verbose	verbosity level (1)
	-V		SONIC Version
	-h, --help	This help message

DAEMON RUNTIME OPTIONS
	-M, --mode	[ thread | rotate ]
	-S, --services	SERVICE_NAME[,SERVICE_NAME,SN]

	-O, --output	[ flush | buffer ] default: flush
	-l, --log	file or path to use for logging

	--s-SERVICE_NAME-OPTION, --service-SN-OPT
	--s-SN-stream	Sonic STREAM_NAME for SN
	--s-SN-max	pthread limit count

	--s-SN-ip	ip address and,
	--s-SN-port	port number to bind service

THREADED ONLY OUTPUT OPTIONS
	--s-SN-thread_out
			service output mode [ flush | buffer ]
			default: daemon runtime default
	--s-SN-thread_buffer_quiet
			dont send service STDOUT when in 'buffer' mode
	--s-SN-thread_log
			log for service only file or path to use

SOCKET OPTIONS
	--s-SN-socket_domain
			ie. AF_INET, AF_INET6, AF_UNIX
	--s-SN-socket_type
			eg. SOCK_STREAM
	--s-SN-socket_protocol
			eg. SO_TCP or SO_UDP
	--s-SN-socket_options
			scalar list, eg. SO_REUSEADDR=1,SO_BROADCAST=1

STREAM / SERVICE OPTIONS
	--service-SERVICE_NAME-STREAM_OPTION
			NB, check relevant stream docs
	--s-SN-byte_size
			max byte size for socket read/write
	--s-SN-verbose
			stream verbosity override

EOF;
  exit();
}

//mode checks
$_sd_ini_file = $SONIC_ARG['i'] ? $SONIC_ARG['i'] : ($SONIC_ARG['ini'] ? $SONIC_ARG['ini'] : $_SERVER['SONIC_ETC'].'/sonic.ini');

$quiet = $SONIC_ARG['q'] ? $SONIC_ARG['q'] : $SONIC_ARG['quiet'];

if ($quiet)
  $verbose = 0;
else
{
  $verbose = $SONIC_ARG['v'] ? $SONIC_ARG['v'] : ($SONIC_ARG['verbose'] ? $SONIC_ARG['verbose'] : NULL);

  if ($verbose != NULL)
    $verbose += 0;
  else
    $verbose = 1;
}

//version
if ($quiet);
else
{
  echo 'SONIC Version (lib.) '.SONIC_VERSION."\n";
  echo 'SONIC Daemon Version '.SONICD_VERSION."\n";
}

if ($SONIC_ARG['V'])
  exit();

flush();

//read config and setup sonic.inc
global	$_SONIC;

if ($_sd_ini = @sonic_ini($_sd_ini_file, true));
else
  die('! ['.$_sd_pid.'] '.sonic_datef()
     .' [SONICD_MAIN] Error: parsing information file '.$_sd_ini_file."\n"
     );

//INI (no shorthand, eg. no --s- only --service-, etc)
$_sd_config = $_sd_ini['SONIC_DAEMON'];
//$_SONIC['SONIC_DAEMON'] = $_sd_config; //for some debug reason
//$_SONIC['SONIC_DAEMON'] = &$_sd_config; //(NB, not secure) for some debug reason
//var_dump($_sd_config); die();
unset($_sd_ini, $_sd_ini_file);

//RUNTIME (will clobber INI, has shorthand)
$SONIC_ARG['output'] = $SONIC_ARG['O'] ? $SONIC_ARG['O'] : ($SONIC_ARG['output'] ? $SONIC_ARG['output'] : 'flush');
$SONIC_ARG['log'] = $SONIC_ARG['l'] ? $SONIC_ARG['l'] : ($SONIC_ARG['log'] ? $SONIC_ARG['log'] : NULL);
$SONIC_ARG['mode'] = $SONIC_ARG['M'] ? $SONIC_ARG['M'] : ($SONIC_ARG['mode'] ? $SONIC_ARG['mode'] : 'thread');
$SONIC_ARG['services'] = $SONIC_ARG['S'] ? $SONIC_ARG['S'] : ($SONIC_ARG['services'] ? $SONIC_ARG['services'] : NULL);
unset($SONIC_ARG['O'], $SONIC_ARG['l'], $SONIC_ARG['S'], $SONIC_ARG['M']);

//NO DEFINE REQUIRING CODE

if ($SONIC_ARG['mode'])
{
  $SONIC_ARG['mode'] = strtolower($SONIC_ARG['mode']);
  switch ($SONIC_ARG['mode'])
  {
    case 'thread':
    case 'threaded':
      if ($_sd_threaded)
        $SONIC_ARG['mode'] = 'thread';
      else
      {
        //TODO: quiet this msg?
        echo '! Warning: thread mode UNsupported, switching to rotate mode'."\n";
        $SONIC_ARG['mode'] = 'rotate';
      }
    break;

    case 'rotate':
    case 'nonthreaded':
    case 'nothread':
      if ($_sd_threaded && !$quiet)
      {
        echo '! Warning: thread mode supported, but in rotate mode'."\n";
      }
      $SONIC_ARG['mode'] = 'rotate';
    break;

    default:
      die('! Error: unknown mode \''.$SONIC_ARG['mode']."', try 'thread' or 'rotate'\n");
  }
} //these wont be reached see set above
else if ($_sd_threaded)
{
  $SONIC_ARG['mode'] = 'thread';
}
else
  $SONIC_ARG['mode'] = 'rotate';
$_sd_config['mode'] = $SONIC_ARG['mode'];

if ($SONIC_ARG['services'])
  $_sd_config['services'] = $SONIC_ARG['services'];
else
  unset($SONIC_ARG['services']);

foreach ($SONIC_ARG as $_s_arg => $_s_aval)
{
  if (preg_match('/^(?:service|s)-([^-]+)-(.*)$/i', $_s_arg, $_s_matches))
  {
    if ($_s_matches[0] = 's')
    {
      if ($SONIC_ARG['service-'.$_s_matches[1].'-'.$_s_matches[2]]);
      else
        $SONIC_ARG['service-'.$_s_matches[1].'-'.$_s_matches[2]] = $_s_aval;

      unset($SONIC_ARG[$_s_arg]);
    }
  }
  else
    continue;

  $_sd_config['service-'.$_s_matches[1].'-'.$_s_matches[2]] = $_s_aval;
}

//SONIC (MAIN/INC - not daemon) SETUP
//global	$_SONIC;
if ($_sonic_ini = $_SONIC['SONIC_INI'])
{
  if ($_sd_config['tmp_path'])
    $_sonic_ini['tmp_path'] = $_sd_config['tmp_path'];

  if ($_sd_config['date_tz'])
    $_sonic_ini['date_tz'] = $_sd_config['date_tz'];

  if ($_sd_config['date_format'])
    $_sonic_ini['date_format'] = $_sd_config['date_format'];

  if ($_sd_config['mime_file'])
    $_sonic_ini['mime_file'] = $_sd_config['mime_file'];

  //php4 socket
  if ($_sd_config['socket_domain'])
    $_sonic_ini['socket_domain'] = $_sd_config['socket_domain'];

  if ($_sd_config['socket_type'])
    $_sonic_ini['socket_type'] = $_sd_config['socket_type'];

  if ($_sd_config['socket_protocol'])
    $_sonic_ini['socket_protocol'] = $_sd_config['socket_protocol'];

  if ($_sd_config['socket_options'])
    $_sonic_ini['socket_options'] = $_sd_config['socket_options'];

  //php5 stream/strm
  if ($_sd_config['strm_transport'])
    $_sonic_ini['strm_transport'] = $_sd_config['strm_stransport'];

  if ($_sd_config['strm_svrflags'])
    $_sonic_ini['strm_svrflags'] = $_sd_config['strm_svrflags'];

  if ($_sd_config['strm_options'])
    $_sonic_ini['strm_options'] = $_sd_config['socket_options'];
}
else
{
  //opt. : '/tmp'); //s_define() has defaults
  $_sonic_ini = array();
  $_sonic_ini['tmp_path'] = ($_sd_config['tmp_path'] ? $_sd_config['tmp_path'] : NULL); 
  $_sonic_ini['date_tz'] = ($_sd_config['date_tz'] ? $_sd_config['date_tz'] : NULL);
  $_sonic_ini['date_format'] = ($_sd_config['date_format'] ? $_sd_config['date_format'] : NULL);
  $_sonic_ini['mime_file'] = ($_sd_config['mime_file'] ? $_sd_config['mime_file'] : NULL);
  $_sonic_ini['socket_domain'] = ($_sd_config['socket_domain'] ? $_sd_config['socket_domain'] : NULL);
  $_sonic_ini['socket_type'] = ($_sd_config['socket_type'] ? $_sd_config['socket_type'] : NULL);
  $_sonic_ini['socket_protocol'] = ($_sd_config['socket_protocol'] ? $_sd_config['socket_protocol'] : NULL);
  $_sonic_ini['socket_options'] = ($_sd_config['socket_options'] ? $_sd_config['socket_options'] : NULL);
  $_sonic_ini['strm_transport'] = ($_sd_config['strm_transport'] ? $_sd_config['strm_transport'] : NULL);
  $_sonic_ini['strm_svrflags'] = ($_sd_config['strm_svrflags'] ? $_sd_config['strm_svrflags'] : NULL);
  $_sonic_ini['strm_options'] = ($_sd_config['strm_options'] ? $_sd_config['strm_options'] : NULL);
}
//GET 'N SET defaults, defines
$_sonic_ini = sonic_define($_sonic_ini);
//
$_sd_config['tmp_path']		= $_sonic_ini['tmp_path'];
$_sd_config['date_tz']		= $_sonic_ini['date_tz'];
$_sd_config['date_format']	= $_sonic_ini['date_format'];
$_sd_config['mime_file']	= $_sonic_ini['mime_file'];
$_sd_config['socket_domain']	= $_sonic_ini['socket_domain'];
$_sd_config['socket_type']	= $_sonic_ini['socket_type'];
$_sd_config['socket_protocol']	= $_sonic_ini['socket_protocol'];
$_sd_config['socket_options']	= $_sonic_ini['socket_options'];
$_sd_config['strm_transport']	= $_sonic_ini['strm_transport'];
$_sd_config['strm_svrflags']	= $_sonic_ini['strm_svrflags'];
$_sd_config['strm_options']	= $_sonic_ini['strm_options'];
unset($_sonic_ini);

//DEFINE REQUIRING CODE

if ($SONIC_ARG['log'])
{
  if ($SONIC_ARG['log']{strlen($SONIC_ARG['log']) - 1} == DIRECTORY_SEPARATOR)
    $SONIC_ARG['log'] = substr($SONIC_ARG['log'], 0, -1);

  if (file_exists($SONIC_ARG['log']))
  {
    //echo append notice
    if (is_dir($SONIC_ARG['log']))
    {
      $SONIC_ARG['log'] .= '/sonic-daemon-'.time().'.log';

      echo '$ Notice: output logfile -> \''.$SONIC_ARG['log'].'\''."\n";
    }
  }

  //write permission test or die
  if (sonic_file_mode_contents($SONIC_ARG['log'], '', 'a')); //has default sonic define pre-req.
  else
    die('! Error: unable to write logfile -> \''.$SONIC_ARG['log'].'\''."\n");

  $_sd_config['log'] = $SONIC_ARG['log'];
}

if ($SONIC_ARG['output'])
{
  $SONIC_ARG['output'] = strtolower($SONIC_ARG['output']);
  switch ($SONIC_ARG['output'])
  {
    case 'flush':
    case 'implicit':
    case 'default':
    case 'normal':
    case 'standard':
      if ($SONIC_ARG['log'])
      {
        //TODO: quiet this msg?
        echo '! Warning: switching modes, write to log file only possible when \'output\' mode is set to \'buffer\''."\n";
        echo '$ Notice: if you want to write to log in \'flush\' mode, use shell output redirection'."\n";
        echo '- ------> eg. '."\t".'`runtime_command.exec > output_log_file.txt`'."\n";

        $SONIC_ARG['output'] = 'buffer';
      }
    break;

    case 'buffer':
    case 'buff':
    case 'log':
    case 'cache':
    break;

    default:
      die('! Error: unknown output mode \''.$SONIC_ARG['output']."', try 'flush' or 'buffer'\n");
  }
} //wont be reached see set above
else
  $SONIC_ARG['output'] = 'flush';

$_sd_config['output'] = $SONIC_ARG['output'];

if ($_sd_config['log'] && $quiet)
{
  $verbose = $SONIC_ARG['v'] ? $SONIC_ARG['v'] : ($SONIC_ARG['verbose'] ? $SONIC_ARG['verbose'] : NULL);

  if ($verbose != NULL)
    $verbose += 0;
  else
    $verbose = 1;

  $quiet = false;

  $_sd_config['buffer_quiet'] = true;

  echo '$ Notice: STDOUT mode set as \'buffer\' with quiet enabled, verbosity reset to \''.$verbose.'\' for LOG buffer'."\n";
}

//MISC
//error_reporting(E_ALL);
//Allow the script to hang around waiting for connections
if (ini_get('max_execution_time') + 0 > 0)
{
  if (ini_get('safe_mode') == '')
    set_time_limit(0);
  else
    echo '! Warning: server running in safe mode with \'max_execution_time\' = \''
        .ini_get('max_execution_time').'\''."\n";
}

//Turn on implicit output flushing so we see what we're getting as it comes in
flush();
ob_implicit_flush();
if ($_sd_config['output'] == 'buffer')
  ob_start();

//TZ, actually set in sonic_define()
if ($quiet);
else
  echo '* timezone ['.SONIC_DATE_TZ.']'."\n";

//IPC SETUP
if ($_sd_ipc_service_config = sonic_daemon_load_config_IPC($_sd_config, array('fatal' => true, 'quiet' => $quiet)))
{
  global	$_sd_ipc_service_info;

  sonic_daemon_load_IPC($_sd_ipc_service_config, array('fatal' => true, 'quiet' => $quiet));

  $_sd_ipc_service_info = array(
			'wait_time' => &$_sd_ipc_service_config['wait_time'],
			'handlers' => &$_sd_ipc_service_config['handlers'],
			'handlers_array' => &$_sd_ipc_service_config['handlers_array'],
			'socket_domain' => &$_sd_ipc_service_config['socket_domain'],
			'socket_type' => &$_sd_ipc_service_config['socket_type'],
			'socket_protocol' => &$_sd_ipc_service_config['socket_protocol'],
			'port' => &$_sd_ipc_service_config['port'],
			'ip' => &$_sd_ipc_service_config['ip'],
			);

  $_sd_is_wait = $_sd_ipc_service_config['wait_time'];
  $_sd_is_wait_time = time() + $_sd_is_wait;
  $_sd_is_endpoint_connected = false;
  $_sd_is_buf = $__sd_is_buf = '';
  $_sd_is_out_default = $_sd_ipc_service_config['default_msg'];
  $_sd_IPC_name = $_sd_ipc_service_config['handlers_array'][0];
  $_sd_is_fin = false;
  $_sd_is_out = $_sd_is_out_default;
}

if ($quiet);
else
  echo '* '.(($SONIC_ARG['mode'] == 'thread') ? '' : 'non ').'threaded server running on pid ['.getmypid()."]\n";

//SELECT and load SERVICES
if (gettype($_sd_config['services']) == 'string');
else
  $_sd_config['services'] = '';

if ($_sd_ipc_service_config)
  $_sd_lso = array('fatal' => false, 'quiet' => $quiet);
else
  $_sd_lso = array('fatal' => true, 'quiet' => $quiet);

list($_sd_services, $_sd_threads_total) = sonic_daemon_load_config_services($_sd_config, $_sd_lso);
//var_dump($_sd_services, $_sd_threads_total);

//create socket for listen service(s)
$_sd_ls = sonic_daemon_load_services($_sd_services, $_sd_lso);
//var_dump($_sd_services, $_sd_ls);
unset($_sd_ls, $_sd_lso);

//daemon main loop
$_sd_config['cycle_rotate_wait_usec'] += 0;
$_sd_config['cycle_thread_wait_time'] += 0;
$_sd_config['cycle_thread_wait_time_IPC_enable'] += 0;
$_sd_config['cycle_thread_wait_usec_IPC_active'] += 0;

if ($_sd_config['cycle_rotate_wait_usec'] <= 0)
  $_sd_config['cycle_rotate_wait_usec'] = 150;

if ($_sd_config['cycle_thread_wait_time'] <= 0)
  $_sd_config['cycle_thread_wait_time']  = 5;

if ($_sd_config['cycle_thread_wait_time_IPC_enable'] <= 0)
  $_sd_config['cycle_thread_wait_time_IPC_enable']  = 1;

if ($_sd_config['cycle_thread_wait_usec_IPC_active'] <= 0)
  $_sd_config['cycle_thread_wait_usec_IPC_active'] = 250;

do
{
  if ($_sd_config['output'] == 'buffer')
  {
    //fork buff safety clear
    $out = ob_get_contents();
    ob_end_clean();

    if (!$_sd_config['buffer_quiet'])
      echo $out;

    if ($_sd_config['log'] && $out != '')
    {
      sonic_file_mode_contents($_sd_config['log'], $out, 'a');

      if ($verbose > 9998)
       echo '= ['.$_sd_pid.'] '.sonic_datef()
            .' [SONICD] lock safe write to log -> \''.$_sd_config['log'].'\''."\n";
    }

    unset($out);

    if ($verbose > 998)
    {
      //before ob_start() so it wont get logged via ob_get_contents()
      echo '|'; //buffer cycle tick
      flush();
    }

    ob_start();
  }
  else if ($verbose > 998)
  {
    echo '|'; //buffer cycle tick
    flush();
  }

  if ($_sd_threaded)
  {
    if ($_sd_threads_shutdown > 0 && $_sd_threads_total > 0)
    {
      if ($verbose > 1)
      {
        echo '$ ['.$_sd_pid.'] '.sonic_datef()
             .' [SONICD_THREADED] Notice: services CLEARED'
             ."\n";
      }

//        var_dump($_sd_services);
//        var_dump($_sd_threads_service);

      $_sd_services = array();

      while ($_sd_daemon_thread_handler_lock) { usleep(250); }
      $_sd_daemon_thread_handler_lock = true;

      $_sd_threads_total = 0;
      $_sd_threads_service_active = array();

      $_sd_daemon_thread_handler_lock = false;

      $_zombies = sonic_daemon_thread_info(array('quiet' => $quiet, 'verbose' => $verbose));
      unset($_zombies);
    }

    if ($_sd_threads_shutdown > 0)
      $_sd_tk = sonic_daemon_thread_kill($_sd_threads, array('quiet' => $quiet, 'verbose' => $verbose));

    if ($_sd_threads_shutdown > 4)
    {
      if ($quiet);
      else
      {
        echo '$ ['.$_sd_pid.'] '.sonic_datef()
            .' [SONICD_THREADED] Notice: signal status from remaining services '
            .($_sd_tk ? 'OK' : 'FAIL')
            ."\n";
      }

      if (sizeof($_sd_threads) > 0)
      {
        $saps = '';
        foreach(array_keys($_sd_threads) as $sapid)
          $saps .= ', '.$sapid;
        $saps = substr($saps, 2);

        if ($verbose > 1)
        {
          echo '$ ['.$_sd_pid.'] '.sonic_datef()
              .' [SONICD_THREADED] Notice: probably zombie but possibly alive -> ['.$saps.']'
              ."\n";
        }
      }

      if ($verbose > 2)
        echo '! ['.$_sd_pid.'] '.sonic_datef()
            .' [SONICD_THREADED] Warning: too many shutdown attempts, exit FORCED'
            ."\n";

      if ($_sd_ipc_service_config)
        sonic_daemon_IPC_socket_shutdown($_sd_ipc_service_config, $opts = array('quiet' => $quiet));

      if (SONICD_SYSV)
        _sonic_daemon_IPC_session_shmsem_unregister_all();

      if ($_sd_config['output'] == 'buffer')
      { 
        //fork buff safety clear
        $out = ob_get_contents();
        ob_end_clean();

        if (!$_sd_config['buffer_quiet'])
          echo $out;

        if ($_sd_config['log'] && $out != '')
          sonic_file_mode_contents($_sd_config['log'], $out, 'a');

        unset($out);
      }

      exit();
    }
    else
      unset($_sd_tk);
  }

  //reset($_sd_services); //reload each
  while (list($service, $service_config) = each($_sd_services))
  {
    if ($SONIC_ARG['mode'] == 'thread')
    {
//         echo	 sizeof($_sd_threads)." < $_sd_threads_total && "
//		.sizeof($_sd_threads_service_active[$service])." < ".$_sd_services[$service]["max"]
//		."\n";

      if (sizeof($_sd_threads_service_active[$service]) < $_sd_services[$service]['max'])
      {
        $_sd_strm_obj = $_sd_strm_obj_name = NULL; //cache destroy
        $_sd_strm_obj_name = 'SONIC_STREAM_'.$service_config['stream'];

        if (class_exists($_sd_strm_obj_name));
        else
        {
          unset($_sd_services[$service]);

          if ($quiet);
          else
            echo '! ['.$_sd_pid.'] '.sonic_datef()
                    .' [SONICD] Error: main() class \''.$_sd_strm_obj_name.'\' does not exists: ['.$service.'] SKIPPED'."\n";
          continue;
        }

        $_sd_strm_obj = new $_sd_strm_obj_name($service_config);

        if (method_exists($_sd_strm_obj, 'sonic_stream_main'));
        else
        {
          unset($_sd_services[$service]);

          if ($quiet);
          else
            echo '! ['.$_sd_pid.'] '.sonic_datef()
                .' [SONICD] Error: main() method \''.$_sd_strm_obj_name.'::sonic_stream_main\' does not exists: ['.$service.'] SKIPPED'."\n";

          continue;
        }

        //daemon core pre-req.
        //careful of stream overwrite init via constructors
        if ($_sd_strm_obj->verbose == NULL)
          $_sd_strm_obj->verbose = $verbose;

        if ($_sd_strm_obj->thread_out == NULL)
          $_sd_strm_obj->thread_out = $_sd_config['output'];

        if ($_sd_strm_obj->thread_buffer_quiet == NULL)
          $_sd_strm_obj->thread_buffer_quiet = $_sd_config['buffer_quiet']; 

        //assignment outside of stream constructor
        //eg. skips sanity check twice (done above already)
        if ($_sd_strm_obj->thread_log == NULL)
          $_sd_strm_obj->thread_log = $_sd_config['log'];

        if ($quiet) {}
        else if ($_sd_services[$service]['ip'])
        {
          echo '+ ['.$_sd_pid.'] '
              .sonic_datef()
              .' ['.$service.'] service listening on '.$_sd_services[$service]['ip'].':'
              .$_sd_services[$service]['port']."\n"
              .'';
        }
        else
        {
          echo '+ ['.$_sd_pid.'] '.sonic_datef().' ['.$service.'] service running'."\n";
        }

        if (SONICD_SYSV && $_sd_strm_obj->key && ($_sd_ipc_sessions['_sonic_daemon_IPC_session_shmsem_register'][$_sd_strm_obj->key] != true))
          _sonic_daemon_IPC_session_shmsem_register($_sd_strm_obj->key);

        $_sd_strm_obj->sonic_stream_thread_main();

	//parent thread manager index vars (same thing, just two for indexing)
	//its my ill' strength hairdryer
        while ($_sd_daemon_thread_handler_lock) { usleep(250); }
        $_sd_daemon_thread_handler_lock = true;
        $_sd_threads[$_sd_strm_obj->pid_child] = array(
					'quiet'		=> $quiet,
					'verbose'	=> $verbose,
					'name'		=> $service_config['name'],
					'socket'	=> &$service_config['socket'],
					'socket_domain'	=> $service_config['socket_domain'],
//					'key'		=> $service_config['key'],
					); 
        $_sd_threads_service[$service][$_sd_strm_obj->pid_child] = &$_sd_threads[$_sd_strm_obj->pid_child];
        $_sd_threads_service_active[$service][$_sd_strm_obj->pid_child] = &$_sd_threads[$_sd_strm_obj->pid_child];
        $_sd_daemon_thread_handler_lock = false;
      }
    }
    else //rotate mode non forked parent - one service at a time
    {
      //pseudo parent->fork pid
      $_sd_strm_obj_pid = $_sd_pid;

      $_sd_strm_obj = $_sd_strm_obj_name = NULL; //cache destroy
      $_sd_strm_obj_name = 'SONIC_STREAM_'.$service_config['stream'];

      if (class_exists($_sd_strm_obj_name));
      else
      {
        unset($_sd_services[$service]);

        if ($quiet);
        else
          echo '! ['.$_sd_pid.'] '.sonic_datef()
              .' [SONICD] Error: main() class \''.$_sd_strm_obj_name.'\' does not exists: ['.$service.'] SKIPPED'."\n";

        continue;
      }

      $_sd_strm_obj = new $_sd_strm_obj_name($service_config);

      if (method_exists($_sd_strm_obj, 'sonic_stream_main'));
      else
      { 
        unset($_sd_services[$service]);

        if ($quiet);
        else
          echo '! ['.$_sd_pid.'] '.sonic_datef()
              .' [SONICD] Error: main() method \''.$_sd_strm_obj_name.'::sonic_stream_main\' does not exists: ['.$service.'] SKIPPED'."\n";

        continue;
      }

      //daemon core pre-req.
      //careful of stream overwrite init via constructors
      if ($_sd_strm_obj->verbose == NULL)
        $_sd_strm_obj->verbose = $verbose;

//      $_sd_strm_obj->sonic_stream_main();

      if ($_sd_threaded)
      {
        //stream rotate mode definition
        $_sd_strm_obj->pid = $_sd_strm_obj_pid;
        $_sd_strm_obj->pid_child = $_sd_strm_obj_pid;

        while ($_sd_daemon_thread_handler_lock) { usleep(250); }
        $_sd_daemon_thread_handler_lock = true;
        $_sd_threads[$_sd_strm_obj_pid] = array(
                                        'quiet'		=> $quiet,
                                        'verbose'	=> $verbose,
                                        'name'		=> $service_config['name'],
                                        'socket'	=> &$service_config['socket'],
					'socket_domain'	=> $service_config['socket_domain'],
//					'key'		=> $service_config['key'],
                                        );
        $_sd_threads_service[$service][$_sd_strm_obj_pid] = &$_sd_threads[$_sd_strm_obj->pid_child];
        $_sd_daemon_thread_handler_lock = false;
      }

      if ($quiet) {}
      else if ($_sd_services[$service]['ip'])
      {
        echo '+ ['.$_sd_strm_obj_pid.'] '.sonic_datef()
            .' ['.$service.'] service listening on '.$_sd_services[$service]['ip'].':'
            .$_sd_services[$service]['port']."\n";
      }
      else
      {
        echo '+ ['.$_sd_strm_obj_pid.'] '.sonic_datef().' ['.$service.'] service running'."\n";
      }

      if (SONICD_SYSV && $_sd_strm_obj->key && ($_sd_ipc_sessions['_sonic_daemon_IPC_session_shmsem_register'][$_sd_strm_obj->key] != true))
        _sonic_daemon_IPC_session_shmsem_register($_sd_strm_obj->key);

      $_sd_strm_obj->sonic_stream_main();

      if ($quiet);
      else
      {
        echo '= ['.$_sd_strm_obj_pid.'] '.sonic_datef().' ['.$_sd_strm_obj->name.'] Exit'."\n";
      }

      if ($_sd_threaded)
      {
        while ($_sd_daemon_thread_handler_lock) { usleep(250); }
        $_sd_daemon_thread_handler_lock = true;
        unset($_sd_threads[$_sd_strm_obj_pid]);
        unset($_sd_threads_service[$service][$_sd_strm_obj_pid]);
        $_sd_daemon_thread_handler_lock = false;
      }

      unset($_sd_strm_obj_pid);
    }

//TODO: if run once remove from services
    if ($service_config['run_once'] && 
        !($SONIC_ARG['mode'] == 'thread' && (sizeof($_sd_threads_service_active[$service]) < $_sd_services[$service]['max'])))
    {
      if ($_sd_threaded)
      {
        while ($_sd_daemon_thread_handler_lock) { usleep(250); }
        $_sd_daemon_thread_handler_lock = true;
        unset($_sd_threads_service_active[$service_config['name']]);
        $_sd_threads_total -= $service_config['max'];
        $_sd_daemon_thread_handler_lock = false;
      }
      unset($_sd_services[$service_config['name']]);
    }
//TODO end

    if ($_sd_threads_shutdown > 0)
      break;
  }
  reset($_sd_services); //reload each

  //IPC loop
  do
  {
    if ($_sd_ipc_service_config['socket']);
    else
      break;

    $_sd_is_read = array($_sd_ipc_service_config['socket']);
    $_sd_is_write = NULL;
    $_sd_is_except = NULL;
    //Warning: socket_select() unable to select [4]: Interrupted system call
    //can be interrupted by thread sig handler
    $_sd_is_changed_sockets = @socket_select($_sd_is_read, $_sd_is_write, $_sd_is_except, 0, 1);

    if (($_sd_is_changed_sockets > 0 OR $_sd_is_changed_sockets != false)
     && ($_sd_ipc_service_config['socket_type'] == SOCK_STREAM && $_sd_is_endpoint_connected == false))
    {
      //socket_accept() will hang until connect - ie. also no need for sleep()
      //0 ok but nothing
      if (($_sd_ipc_service_config['socket_live'] = @socket_accept($_sd_ipc_service_config['socket'])) == false)
        die('! ['.$_sd_pid.'] '.sonic_datef()
           .' [SONICD_IPC] Error: socket_accept(): '
           .socket_strerror(socket_last_error())."\n"
           );

      $_sd_ipc_service_config['socket_listen'] = $_sd_ipc_service_config['socket'];
      $_sd_ipc_service_config['socket'] = &$_sd_ipc_service_config['socket_live'];

      $_sd_is_endpoint_connected = true;

      $_sd_is_wait_time = time() + $_sd_is_wait;
    }

    if (($_sd_is_changed_sockets > 0 OR $_sd_is_changed_sockets != false) OR $_sd_ipc_service_config['socket_type'] != SOCK_STREAM)
    {
      $_sd_is_sopt_rt = socket_get_option($_sd_ipc_service_config['socket'], SOL_SOCKET, SO_RCVTIMEO);
      socket_set_option($_sd_ipc_service_config['socket'], SOL_SOCKET, SO_RCVTIMEO, array('sec' => $_sd_is_wait, 'usec' => 0));

      if (false == ($__sd_is_buf = socket_read($_sd_ipc_service_config['socket'], SONIC_IPC_BYTE_SIZE, PHP_BINARY_READ)))
      {
        if ($_sd_ipc_service_config['socket_type'] == SOCK_STREAM)
          socket_write($_sd_ipc_service_config['socket'], $_sd_is_out, strlen($_sd_is_out));
        else
          echo $_sd_is_out;

        if ($_sd_ipc_service_config['socket_domain'] == AF_UNIX)
        {
          $remote_ip = $_sd_ipc_service_config['ip'];
          $remote_port = $_sd_ipc_service_config['port'];
        }
        else
          @socket_getpeername($_sd_ipc_service_config['socket'], $remote_ip, $remote_port);

        $remote_port += 0;

        if ($quiet);
        else
          echo '+ ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] closed connection '.$remote_ip.':'.$remote_port."\n";

        unset($remote_port, $remote_ip);

        @socket_close($_sd_ipc_service_config['socket']);
 
        if ($_sd_is_endpoint_connected) // && $_sd_ipc_service_config['socket_type'] == SOCK_STREAM)
        {
          unset($_sd_ipc_service_config['socket_live']);
          $_sd_ipc_service_config['socket'] = $_sd_ipc_service_config['socket_listen'];
          unset($_sd_ipc_service_config['socket_listen']);

          $_sd_is_endpoint_connected = false;
        }

        $_sd_is_wait_time = time() + $_sd_is_wait;
        $_sd_is_buf = $__sd_is_buf = '';
        $_sd_is_out = $_sd_is_out_default;
        $_sd_is_fin = false;

        break;
      }

//      else if ($verbose > 998 OR $verbose == 22)
//      {
//        echo $__sd_is_buf;
//      }

      socket_set_option($_sd_ipc_service_config['socket'], SOL_SOCKET, SO_RCVTIMEO, $_sd_is_sopt_rt);

      if ($__sd_is_buf == '')
        break;

      $_sd_is_buf .= $__sd_is_buf;

//IPC MAIN
      $_sd_services_reset = false;
      $_sd_IPC_opts = array(
		'argv'		=> $SONIC_ARG,
		'quiet'		=> $quiet,
		'verbose'	=> $verbose,
		'config'	=> &$_sd_config,
		'services'	=> $_sd_services,
		'buffer'	=> $_sd_is_buf,
		'finished'	=> &$_sd_is_fin,
		'output'	=> &$_sd_is_out,
		'services_reset'=> &$_sd_services_reset
		);

      call_user_func_array('sonic_daemon_IPC_'.$_sd_IPC_name.'_main', array(&$_sd_IPC_opts));
      unset($_sd_IPC_opts);

      if ($_sd_services_reset)
      {
        if ($_sd_threaded && $_sd_threads_shutdown > 0)
        {
          $_sd_threads_shutdown = 0;
          if ($quiet);
          else
            echo '$ ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] Notice: main(): service shutdown STOPPED'."\n";
        }

        if ($quiet);
        else
          echo '$ ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] Notice: main(): service information RESET'."\n";

        list($_sd_services_new, $_sd_threads_total) = sonic_daemon_load_config_services($_sd_config, array('fatal' => false, 'quiet' => $quiet, 'prelog' => true));

        if ($SONIC_ARG['mode'] == 'thread')
          $_sd_rst_service = $_sd_threads_service_active;
        else
          $_sd_rst_service = array($_sd_strm_obj_name => array());

        foreach ($_sd_rst_service as $_sts_rst_name => $_sts_rst_pids)
        {
          //kill the kids, zombies might still exist
          if ($SONIC_ARG['mode'] == 'thread')
          {
            foreach ($_sts_rst_pids as $_pid => $_info)
            {
//echo 'sending int to ('.$_pid.')'."\n";
              posix_kill($_pid, SIGINT); //SIGQUIT, SIGTERM
//echo 'waiting for ('.$_pid.')'."\n";
              usleep(150000);
            }

            $time_start = time();
            $time_limit = 8; //seconds each thread
            $si_x = 2;

            foreach ($_sts_rst_pids as $_pid => $_info)
            {
              //SIGSTOP or SIGKILL will cause non graceful shutdown
              $i = 0;
              $j = 1;
              $k = true;
              $l = true;
              while ($l && $k && $_sd_services_new[$_info['name']])
              {
                $i++;

                if ($_sd_threads[$_pid])
                {
                  if (($time_start + $time_limit) > time())
                  {
//echo 'still waiting for ('.$_pid.')'."\n";

                    usleep(150000);
                  }
                  else if ($j > $si_x)
                  {
                    if ($quiet);
                    else
                      echo '$ ['.$_sd_pid.'] '.sonic_datef()
                          .' [SONICD_IPC] Notice: main(): ['.$_pid.'] zombified by '.$si_x.' signals x time(s) wait of '.$time_limit.' seconds'."\n";

//echo 'issue kill or report zombie error'."\n";
//echo 'keep old service, threads will repopulate, or throw skip warning and remove service (zombies will remain regardless)'."\n";
                    $_time_start = time();
                    $_possible_recover = false;

                    while (($_time_start + $time_limit) > time())
                    {
                      posix_kill($_pid, SIGKILL);

                      usleep(150000);

                      if ($_sd_threads[$_pid])
                        usleep(150000);
                      else
                        $_possible_recover = true;
                    }
                    unset($_time_start);

                    if ($_possible_recover)
                    {
                      if ($quiet);
                      else
                        echo '! ['.$_sd_pid.'] '.sonic_datef()
                            .' [SONICD_IPC] Warning: main(): ['.$_pid.'] was non cleanly KILLED'."\n";

//                      $k = true;
                      break;
                    }
                    else if ($_sd_services[$_info['name']]['socket'] && socket_get_option($_sd_services[$_info['name']]['socket'], SOL_SOCKET, SO_REUSEADDR) != 1)
                    {
                      unset($_sd_services_new[$_info['name']]);

                      if ($quiet);
                      else
                        echo '$ ['.$_sd_pid.'] '.sonic_datef()
                            .' [SONICD_IPC] "Notice: main(): ['.$_info['name'].'] SKIPPED'."\n";
                    }
                    unset($_possible_recover);

                    $k = false;
                  }
                  else
                  {
                    $j++;
                    $time_start = time();
//echo 'sending int again to ('.$_pid.')'."\n";
                    posix_kill($_pid, SIGTERM);
//echo 'still waiting after retry for ('.$_pid.')'."\n";
                    usleep(150000);
                  }
                }
                else
                  $l = false;
              }

              if ($k);
              else
                break;
//var_dump($i);
            }
//
          }

          if ($_sd_services[$_sts_rst_name]['socket'])
          {
            if ($quiet);
            else
              echo '$ ['.$_sd_pid.'] '.sonic_datef()
                  .' [SONICD_IPC] Notice: main(): ['.$_sts_rst_name.'] connection shutdown and close'."\n";

            socket_getsockname($_sd_services[$_sts_rst_name]['socket'], $_sts_rst_saddr, $_sts_rst_sport);
//hopefully shutdown already
            @socket_shutdown($_sd_services[$_sts_rst_name]['socket'], 2);
            usleep(250);
//shouldnt throw error but who knows if there was a cleanup
            @socket_close($_sd_services[$_sts_rst_name]['socket']);
            usleep(250);

            unset($_sd_services[$_sts_rst_name]['socket']);

            if ($_sd_services['socket_domain'] == AF_UNIX)
              @unlink($_sts_rst_saddr);

            unset($_sts_rst_saddr, $_sts_rst_sport);
          }
        }
        unset($_sd_rst_service);

        if (SONICD_SYSV)
        {
          foreach ($_sd_services as $_ssdn => $ssdc)
            _sonic_daemon_IPC_session_shmsem_unregister($ssdc['key']);
        }
//end clear

        //create socket for listen service(s)
        $_sd_services_new_bak = $_sd_services_new;
        if (sonic_daemon_load_services($_sd_services_new, array('fatal' => false, 'quiet' => $quiet)));
        else
        {
          foreach ($_sd_services_new_bak as $_ssn_s => $_ssn_sc)
          {
            if ($_sd_services_new[$_ssn_s]);
            else
            {
              $_sd_threads_total -= $_sd_services_new_bak[$_ssn_s]['max'];
              if ($quiet);
              else
                echo '$ ['.$_sd_pid.'] '.sonic_datef()
                    .' [SONICD_IPC] Notice: main(): ['.$_ssn_s.'] SKIPPED'."\n";
            }
          }
        }
        unset($_sd_services_new_bak);

        $_sd_services = $_sd_services_new;

        unset($_sd_services_new);

        if ($SONIC_ARG['mode'] == 'thread')
        {
//echo 'sleeping'."\n";

          while ($_sd_daemon_thread_handler_lock) { usleep(250); }
          $_sd_daemon_thread_handler_lock = true;
          $_sd_threads_service_active = array();
          $_sd_daemon_thread_handler_lock = false;
          $_zombies = sonic_daemon_thread_info(array('quiet' => $quiet, 'verbose' => $verbose));
          unset($_zombies);
        }
      }
 
//IPC MAIN

//      socket_set_option($_sd_ipc_service_config['socket'], SOL_SOCKET, SO_RCVTIMEO, $_sd_is_sopt_rt);

      //TODO: optional update waittime, but can introduce timed ddos block?
      continue;
    }

    break;

  } while (true);

  if ($_sd_ipc_service_config['socket'])
  {
    $_sd_is_time_check = time() < $_sd_is_wait_time;

    if (socket_last_error($_sd_ipc_service_config['socket']) == 0 && $_sd_is_endpoint_connected == false) {}
    else if ($_sd_is_fin == false
         &&  socket_last_error($_sd_ipc_service_config['socket']) == 0
         && ($_sd_is_endpoint_connected && $_sd_is_time_check)) {}
    else
    {
      if ($_sd_ipc_service_config['socket_type'] == SOCK_STREAM)
        socket_write($_sd_ipc_service_config['socket'], $_sd_is_out, strlen($_sd_is_out));
//TODO, something about this?
//      else
//        echo $_sd_is_out;

      if ($_sd_ipc_service_config['socket_domain'] == AF_UNIX)
      {
        $remote_ip = $_sd_ipc_service_config['ip'];
        $remote_port = $_sd_ipc_service_config['port'];
      }
      else
        @socket_getpeername($_sd_ipc_service_config['socket'], $remote_ip, $remote_port);

      $remote_port += 0;

      if ($quiet);
      else
        echo '+ ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] closed connection '.$remote_ip.':'.$remote_port."\n";

      unset($remote_port, $remote_ip);

      @socket_close($_sd_ipc_service_config['socket']);

      if ($_sd_is_endpoint_connected)
      {
        unset($_sd_ipc_service_config['socket_live']);
        $_sd_ipc_service_config['socket'] = $_sd_ipc_service_config['socket_listen'];
        unset($_sd_ipc_service_config['socket_listen']);

        $_sd_is_endpoint_connected = false;

        if ($_sd_is_time_check) {}
        else if ($quiet) {}
        else
        {
          echo '$ ['.$_sd_pid.'] '.sonic_datef().' [SONICD_IPC] Notice: main(): timeout '.$_sd_is_wait.' seconds'."\n";
        }
      }

      $_sd_is_wait_time = time() + $_sd_is_wait;
      $_sd_is_buf = $__sd_is_buf = '';
      $_sd_is_out = $_sd_is_out_default; 
      $_sd_is_fin = false;
    }
  }
  else if ($_sd_ipc_service_config)
  {
    //TODO: turn this into shutdown?
    echo '! ['.$_sd_pid.'] '.sonic_datef()
        .' [SONICD_IPC] Warning: main(): IPC enabled but no listening socket is loaded...'."\n";
  }

  if ($_sd_threaded)
  {
    if ($_sd_threads_shutdown > 0)
    {
      if ($SONIC_ARG['mode'] == 'thread' && sizeof($_sd_threads) > 0);
      else
      {
        if ($verbose > 1)
          echo '+ ['.$_sd_pid.'] '.sonic_datef().' [SONICD_THREADED] Exit'."\n";

        if ($_sd_ipc_service_config)
          sonic_daemon_IPC_socket_shutdown($_sd_ipc_service_config, $opts = array('quiet' => $quiet));

        if (SONICD_SYSV)
          _sonic_daemon_IPC_session_shmsem_unregister_all();

        if ($_sd_config['output'] == 'buffer')
        {
          //fork buff safety clear
          $out = ob_get_contents();
          ob_end_clean();

          if (!$_sd_config['buffer_quiet'])
            echo $out;

          if ($_sd_config['log'] && $out != '')
            sonic_file_mode_contents($_sd_config['log'], $out, 'a');

          unset($out);
        }

        exit();
      }
    }

    if (sizeof($_sd_threads) < $_sd_threads_total) { }
    else if ($SONIC_ARG['mode'] == 'rotate' && sizeof($_sd_services) > 0) { }
    //conf'd sleep times
    else if ($_sd_is_endpoint_connected)
    {
      usleep($_sd_config['cycle_thread_wait_usec_IPC_active']);
    }
    else if ($_sd_ipc_service_config['socket'])
    {
      $_sd_is_read = array($_sd_ipc_service_config['socket']);
      $_sd_is_write = NULL;
      $_sd_is_except = NULL;
      //Warning: socket_select() unable to select [4]: Interrupted system call
      //can be interrupted by thread sig handler
      $_sd_is_changed_sockets = @socket_select($_sd_is_read, $_sd_is_write, $_sd_is_except, $_sd_config['cycle_thread_wait_time_IPC_enable']);
    }
    else if (($_sd_threaded && $_sd_threads_total < 1 && sizeof($_sd_threads) < 1) OR ($SONIC_ARG['mode'] == 'rotate' ))
    {
      if ($_sd_config['output'] == 'buffer')
      {
        //fork buff safety clear
        $out = ob_get_contents();
        ob_end_clean();

        if (!$_sd_config['buffer_quiet'])
          echo $out;

        if ($_sd_config['log'] && $out != '')
          sonic_file_mode_contents($_sd_config['log'], $out, 'a');

        unset($out);
      }

      exit(0);
    }
    else
    {
      //have a kit-kat
      sleep($_sd_config['cycle_thread_wait_time']);
    }
  }
  else //worst case no pcntl(/thread) rotate mode
    usleep($_sd_config['cycle_rotate_wait_usec']);

} while (true);
//MAIN

//SCOPE
}

sonic_daemon_main();

?>
